#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import os
import unittest
from unittest.mock import patch

from evdev.ecodes import EV_KEY, EV_ABS

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.configs.mapping import UIMapping
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from tests.lib.test_setup import test_setup


@test_setup
class TestPreset(unittest.TestCase):
    def setUp(self):
        self.preset = Preset(PathUtils.get_preset_path("foo", "bar2"))
        self.assertFalse(self.preset.has_unsaved_changes())

    def test_is_mapped_multiple_times(self):
        combination = InputCombination(
            InputCombination.from_tuples((1, 1, 1), (2, 2, 2), (3, 3, 3), (4, 4, 4))
        )
        permutations = combination.get_permutations()
        self.assertEqual(len(permutations), 6)

        self.preset._mappings[permutations[0]] = Mapping(
            input_combination=permutations[0],
            target_uinput="keyboard",
            output_symbol="a",
        )
        self.assertFalse(self.preset._is_mapped_multiple_times(permutations[2]))

        self.preset._mappings[permutations[1]] = Mapping(
            input_combination=permutations[1],
            target_uinput="keyboard",
            output_symbol="a",
        )
        self.assertTrue(self.preset._is_mapped_multiple_times(permutations[2]))

    def test_has_unsaved_changes(self):
        self.preset.path = PathUtils.get_preset_path("foo", "bar2")
        self.preset.add(Mapping.from_combination())
        self.assertTrue(self.preset.has_unsaved_changes())
        self.preset.save()
        self.assertFalse(self.preset.has_unsaved_changes())

        self.preset.empty()
        self.assertEqual(len(self.preset), 0)
        # empty preset but non-empty file
        self.assertTrue(self.preset.has_unsaved_changes())

        # load again from the disc
        self.preset.load()
        self.assertEqual(
            self.preset.get_mapping(InputCombination.empty_combination()),
            Mapping.from_combination(),
        )
        self.assertFalse(self.preset.has_unsaved_changes())

        # change the path to a non exiting file
        self.preset.path = PathUtils.get_preset_path("bar", "foo")
        # the preset has a mapping, the file has not
        self.assertTrue(self.preset.has_unsaved_changes())

        # change back to the original path
        self.preset.path = PathUtils.get_preset_path("foo", "bar2")
        # no difference between file and memory
        self.assertFalse(self.preset.has_unsaved_changes())

        # modify the mapping
        mapping = self.preset.get_mapping(InputCombination.empty_combination())
        mapping.gain = 0.5
        self.assertTrue(self.preset.has_unsaved_changes())
        self.preset.load()

        self.preset.path = PathUtils.get_preset_path("bar", "foo")
        self.preset.remove(Mapping.from_combination().input_combination)
        # empty preset and empty file
        self.assertFalse(self.preset.has_unsaved_changes())

        self.preset.path = PathUtils.get_preset_path("foo", "bar2")
        # empty preset, but non-empty file
        self.assertTrue(self.preset.has_unsaved_changes())
        self.preset.load()
        self.assertEqual(len(self.preset), 1)
        self.assertFalse(self.preset.has_unsaved_changes())

        # delete the preset from the system:
        self.preset.empty()
        self.preset.save()
        self.preset.load()
        self.assertFalse(self.preset.has_unsaved_changes())
        self.assertEqual(len(self.preset), 0)

    def test_save_load(self):
        one = InputConfig(type=EV_KEY, code=10)
        two = InputConfig(type=EV_KEY, code=11)
        three = InputConfig(type=EV_KEY, code=12)

        self.preset.add(
            Mapping.from_combination(InputCombination([one]), "keyboard", "1")
        )
        self.preset.add(
            Mapping.from_combination(InputCombination([two]), "keyboard", "2")
        )
        self.preset.add(
            Mapping.from_combination(InputCombination((two, three)), "keyboard", "3"),
        )
        self.preset.path = PathUtils.get_preset_path("Foo Device", "test")
        self.preset.save()

        path = os.path.join(
            PathUtils.config_path(), "presets", "Foo Device", "test.json"
        )
        self.assertTrue(os.path.exists(path))

        loaded = Preset(PathUtils.get_preset_path("Foo Device", "test"))
        self.assertEqual(len(loaded), 0)
        loaded.load()

        self.assertEqual(len(loaded), 3)
        self.assertRaises(TypeError, loaded.get_mapping, one)
        self.assertEqual(
            loaded.get_mapping(InputCombination([one])),
            Mapping.from_combination(InputCombination([one]), "keyboard", "1"),
        )
        self.assertEqual(
            loaded.get_mapping(InputCombination([two])),
            Mapping.from_combination(InputCombination([two]), "keyboard", "2"),
        )
        self.assertEqual(
            loaded.get_mapping(InputCombination([two, three])),
            Mapping.from_combination(InputCombination([two, three]), "keyboard", "3"),
        )

        # load missing file
        preset = Preset(PathUtils.get_config_path("missing_file.json"))
        self.assertRaises(FileNotFoundError, preset.load)

    def test_modify_mapping(self):
        ev_1 = InputCombination([InputConfig(type=EV_KEY, code=1)])
        ev_3 = InputCombination([InputConfig(type=EV_KEY, code=2)])
        # only values between -99 and 99 are allowed as mapping for EV_ABS or EV_REL
        ev_4 = InputCombination([InputConfig(type=EV_ABS, code=1, analog_threshold=99)])

        # add the first mapping
        self.preset.add(Mapping.from_combination(ev_1, "keyboard", "a"))
        self.assertTrue(self.preset.has_unsaved_changes())
        self.assertEqual(len(self.preset), 1)

        # change ev_1 to ev_3 and change a to b
        mapping = self.preset.get_mapping(ev_1)
        mapping.input_combination = ev_3
        mapping.output_symbol = "b"
        self.assertIsNone(self.preset.get_mapping(ev_1))
        self.assertEqual(
            self.preset.get_mapping(ev_3),
            Mapping.from_combination(ev_3, "keyboard", "b"),
        )
        self.assertEqual(len(self.preset), 1)

        # add 4
        self.preset.add(Mapping.from_combination(ev_4, "keyboard", "c"))
        self.assertEqual(
            self.preset.get_mapping(ev_3),
            Mapping.from_combination(ev_3, "keyboard", "b"),
        )
        self.assertEqual(
            self.preset.get_mapping(ev_4),
            Mapping.from_combination(ev_4, "keyboard", "c"),
        )
        self.assertEqual(len(self.preset), 2)

        # change the preset of 4 to d
        mapping = self.preset.get_mapping(ev_4)
        mapping.output_symbol = "d"
        self.assertEqual(
            self.preset.get_mapping(ev_4),
            Mapping.from_combination(ev_4, "keyboard", "d"),
        )
        self.assertEqual(len(self.preset), 2)

        # try to change combination of 4 to 3
        mapping = self.preset.get_mapping(ev_4)
        with self.assertRaises(KeyError):
            mapping.input_combination = ev_3

        self.assertEqual(
            self.preset.get_mapping(ev_3),
            Mapping.from_combination(ev_3, "keyboard", "b"),
        )
        self.assertEqual(
            self.preset.get_mapping(ev_4),
            Mapping.from_combination(ev_4, "keyboard", "d"),
        )
        self.assertEqual(len(self.preset), 2)

    def test_avoids_redundant_saves(self):
        with patch.object(self.preset, "has_unsaved_changes", lambda: False):
            self.preset.path = PathUtils.get_preset_path("foo", "bar2")
            self.preset.add(Mapping.from_combination())
            self.preset.save()

        with open(PathUtils.get_preset_path("foo", "bar2"), "r") as f:
            content = f.read()

        self.assertFalse(content)

    def test_combinations(self):
        ev_1 = InputConfig(type=EV_KEY, code=1, analog_threshold=111)
        ev_2 = InputConfig(type=EV_KEY, code=1, analog_threshold=222)
        ev_3 = InputConfig(type=EV_KEY, code=2, analog_threshold=111)
        ev_4 = InputConfig(type=EV_ABS, code=1, analog_threshold=99)
        combi_1 = InputCombination((ev_1, ev_2, ev_3))
        combi_2 = InputCombination((ev_2, ev_1, ev_3))
        combi_3 = InputCombination((ev_1, ev_2, ev_4))

        self.preset.add(Mapping.from_combination(combi_1, "keyboard", "a"))
        self.assertEqual(
            self.preset.get_mapping(combi_1),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_2),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )
        # since combi_1 and combi_2 are equivalent, this raises a KeyError
        self.assertRaises(
            KeyError,
            self.preset.add,
            Mapping.from_combination(combi_2, "keyboard", "b"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_1),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_2),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )

        self.preset.add(Mapping.from_combination(combi_3, "keyboard", "c"))
        self.assertEqual(
            self.preset.get_mapping(combi_1),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_2),
            Mapping.from_combination(combi_1, "keyboard", "a"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_3),
            Mapping.from_combination(combi_3, "keyboard", "c"),
        )

        mapping = self.preset.get_mapping(combi_1)
        mapping.output_symbol = "c"
        with self.assertRaises(KeyError):
            mapping.input_combination = combi_3

        self.assertEqual(
            self.preset.get_mapping(combi_1),
            Mapping.from_combination(combi_1, "keyboard", "c"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_2),
            Mapping.from_combination(combi_1, "keyboard", "c"),
        )
        self.assertEqual(
            self.preset.get_mapping(combi_3),
            Mapping.from_combination(combi_3, "keyboard", "c"),
        )

    def test_remove(self):
        # does nothing
        ev_1 = InputCombination([InputConfig(type=EV_KEY, code=40)])
        ev_2 = InputCombination([InputConfig(type=EV_KEY, code=30)])
        ev_3 = InputCombination([InputConfig(type=EV_KEY, code=20)])
        ev_4 = InputCombination([InputConfig(type=EV_KEY, code=10)])

        self.assertRaises(TypeError, self.preset.remove, (EV_KEY, 10, 1))
        self.preset.remove(ev_1)
        self.assertFalse(self.preset.has_unsaved_changes())
        self.assertEqual(len(self.preset), 0)

        self.preset.add(Mapping.from_combination(input_combination=ev_1))
        self.assertEqual(len(self.preset), 1)
        self.preset.remove(ev_1)
        self.assertEqual(len(self.preset), 0)

        self.preset.add(Mapping.from_combination(ev_4, "keyboard", "KEY_KP1"))
        self.assertTrue(self.preset.has_unsaved_changes())
        self.preset.add(Mapping.from_combination(ev_3, "keyboard", "KEY_KP2"))
        self.preset.add(Mapping.from_combination(ev_2, "keyboard", "KEY_KP3"))
        self.assertEqual(len(self.preset), 3)
        self.preset.remove(ev_3)
        self.assertEqual(len(self.preset), 2)
        self.assertEqual(
            self.preset.get_mapping(ev_4),
            Mapping.from_combination(ev_4, "keyboard", "KEY_KP1"),
        )
        self.assertIsNone(self.preset.get_mapping(ev_3))
        self.assertEqual(
            self.preset.get_mapping(ev_2),
            Mapping.from_combination(ev_2, "keyboard", "KEY_KP3"),
        )

    def test_empty(self):
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=10)]),
                "keyboard",
                "1",
            ),
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=11)]),
                "keyboard",
                "2",
            ),
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=12)]),
                "keyboard",
                "3",
            ),
        )
        self.assertEqual(len(self.preset), 3)
        self.preset.path = PathUtils.get_config_path("test.json")
        self.preset.save()
        self.assertFalse(self.preset.has_unsaved_changes())

        self.preset.empty()
        self.assertEqual(self.preset.path, PathUtils.get_config_path("test.json"))
        self.assertTrue(self.preset.has_unsaved_changes())
        self.assertEqual(len(self.preset), 0)

    def test_clear(self):
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=10)]),
                "keyboard",
                "1",
            ),
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=11)]),
                "keyboard",
                "2",
            ),
        )
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=12)]),
                "keyboard",
                "3",
            ),
        )
        self.assertEqual(len(self.preset), 3)
        self.preset.path = PathUtils.get_config_path("test.json")
        self.preset.save()
        self.assertFalse(self.preset.has_unsaved_changes())

        self.preset.clear()
        self.assertFalse(self.preset.has_unsaved_changes())
        self.assertIsNone(self.preset.path)
        self.assertEqual(len(self.preset), 0)

    def test_dangerously_mapped_btn_left(self):
        # btn left is mapped
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig.btn_left()]),
                "keyboard",
                "1",
            )
        )
        self.assertTrue(self.preset.dangerously_mapped_btn_left())
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=41)]),
                "keyboard",
                "2",
            )
        )
        self.assertTrue(self.preset.dangerously_mapped_btn_left())

        # another mapping maps to btn_left
        self.preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=42)]),
                "mouse",
                "btn_left",
            )
        )
        self.assertFalse(self.preset.dangerously_mapped_btn_left())

        mapping = self.preset.get_mapping(
            InputCombination([InputConfig(type=EV_KEY, code=42)])
        )
        mapping.output_symbol = "BTN_Left"
        self.assertFalse(self.preset.dangerously_mapped_btn_left())

        mapping.target_uinput = "keyboard + mouse"
        mapping.output_symbol = "3"
        self.assertTrue(self.preset.dangerously_mapped_btn_left())

        # btn_left is not mapped
        self.preset.remove(InputCombination([InputConfig.btn_left()]))
        self.assertFalse(self.preset.dangerously_mapped_btn_left())

    def test_save_load_with_invalid_mappings(self):
        ui_preset = Preset(
            PathUtils.get_config_path("test.json"), mapping_factory=UIMapping
        )

        ui_preset.add(UIMapping())
        self.assertFalse(ui_preset.is_valid())

        # make the mapping valid
        m = ui_preset.get_mapping(InputCombination.empty_combination())
        m.output_symbol = "a"
        m.target_uinput = "keyboard"
        self.assertTrue(ui_preset.is_valid())

        m2 = UIMapping(
            input_combination=InputCombination([InputConfig(type=1, code=2)])
        )
        ui_preset.add(m2)
        self.assertFalse(ui_preset.is_valid())
        ui_preset.save()

        # only the valid preset is loaded
        preset = Preset(PathUtils.get_config_path("test.json"))
        preset.load()
        self.assertEqual(len(preset), 1)

        a = preset.get_mapping(m.input_combination).dict()
        b = m.dict()
        a.pop("mapping_type")
        b.pop("mapping_type")
        self.assertEqual(a, b)
        # self.assertEqual(preset.get_mapping(m.input_combination), m)

        # both presets load
        ui_preset.clear()
        ui_preset.path = PathUtils.get_config_path("test.json")
        ui_preset.load()
        self.assertEqual(len(ui_preset), 2)

        a = ui_preset.get_mapping(m.input_combination).dict()
        b = m.dict()
        a.pop("mapping_type")
        b.pop("mapping_type")
        self.assertEqual(a, b)
        # self.assertEqual(ui_preset.get_mapping(m.input_combination), m)
        self.assertEqual(ui_preset.get_mapping(m2.input_combination), m2)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev._ecodes import EV_ABS, ABS_X, BTN_WEST, BTN_Y, EV_KEY, KEY_A

from inputremapper.utils import get_evdev_constant_name
from tests.lib.test_setup import test_setup


@test_setup
class TestUtil(unittest.TestCase):
    def test_get_evdev_constant_name(self):
        # BTN_WEST and BTN_Y both are code 308. I don't care which one is chosen
        # in the return value, but it should return one of them without crashing.
        self.assertEqual(get_evdev_constant_name(EV_KEY, BTN_Y), "BTN_WEST")
        self.assertEqual(get_evdev_constant_name(EV_KEY, BTN_WEST), "BTN_WEST")

        self.assertEqual(get_evdev_constant_name(123, KEY_A), "unknown")
        self.assertEqual(get_evdev_constant_name(EV_KEY, 9999), "unknown")

        self.assertEqual(get_evdev_constant_name(EV_KEY, KEY_A), "KEY_A")

        self.assertEqual(get_evdev_constant_name(EV_ABS, ABS_X), "ABS_X")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.task import Task


class HoldTask(Task):
    """Loop the macro until the trigger-key is released."""

    argument_configs = [
        ArgumentConfig(
            name="macro",
            position=0,
            types=[Macro, str, None],
        )
    ]

    async def run(self, callback) -> None:
        macro = self.get_argument("macro").get_value()

        if macro is None:
            await self._trigger_release_event.wait()
            return

        if isinstance(macro, str):
            # if macro is a key name, hold down the key while the
            # keyboard key is physically held down
            symbol = macro
            self.get_argument("macro").assert_is_symbol(symbol)

            code = keyboard_layout.get(symbol)
            callback(EV_KEY, code, 1)
            await self._trigger_release_event.wait()
            callback(EV_KEY, code, 0)

        if isinstance(macro, Macro):
            # repeat the macro forever while the key is held down
            while self.is_holding():
                # run the child macro completely to avoid
                # not-releasing any key
                await macro.run(callback)
                # give some other code a chance to run
                await asyncio.sleep(1 / 1000)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import atexit
import multiprocessing
import os
import time
import unittest
from contextlib import contextmanager
from typing import Tuple, List, Optional
from unittest.mock import patch

import evdev
import gi
import sys

from inputremapper.injection.global_uinputs import GlobalUInputs, FrontendUInput, UInput
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from tests.lib.cleanup import cleanup
from tests.lib.constants import EVENT_READ_TIMEOUT
from tests.lib.fixtures import prepare_presets
from tests.lib.logger import logger

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "4")
gi.require_version("GLib", "2.0")
from gi.repository import Gtk, GLib, GtkSource

from inputremapper.configs.mapping import Mapping
from inputremapper.configs.global_config import GlobalConfig
from inputremapper.groups import _Groups
from inputremapper.gui.data_manager import DataManager
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
)
from inputremapper.gui.components.editor import (
    MappingSelectionLabel,
)
from inputremapper.gui.controller import Controller
from inputremapper.gui.reader_service import ReaderService
from inputremapper.gui.utils import gtk_iteration
from inputremapper.gui.user_interface import UserInterface
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.daemon import Daemon, DaemonProxy
from inputremapper.bin.input_remapper_gtk import InputRemapperGtkBin


# iterate a few times when Gtk.main() is called, but don't block
# there and just continue to the tests while the UI becomes
# unresponsive
Gtk.main = gtk_iteration

# doesn't do much except avoid some Gtk assertion error, whatever:
Gtk.main_quit = lambda: None


def launch() -> Tuple[
    UserInterface,
    Controller,
    DataManager,
    MessageBroker,
    DaemonProxy,
    GlobalConfig,
]:
    """Start input-remapper-gtk."""
    with patch.object(sys, "argv", ["/usr/bin/input-remapper-gtk", "-d"]):
        return_ = InputRemapperGtkBin.main()

    gtk_iteration()
    # otherwise a new handler is added with each call to launch, which
    # spams tons of garbage when all tests finish
    atexit.unregister(InputRemapperGtkBin.stop)
    return return_


def start_reader_service():
    def process():
        global_uinputs = GlobalUInputs(FrontendUInput)
        reader_service = ReaderService(_Groups(), global_uinputs)
        loop = asyncio.new_event_loop()
        loop.run_until_complete(reader_service.run())

    multiprocessing.Process(target=process).start()


def os_system_patch(cmd, original_os_system=os.system):
    # instead of running pkexec, fork instead. This will make
    # the reader-service aware of all the test patches
    if "pkexec input-remapper-control --command start-reader-service" in cmd:
        logger.info("pkexec-patch starting ReaderService process")
        start_reader_service()
        return 0

    return original_os_system(cmd)


@contextmanager
def patch_services():
    """Don't connect to the dbus and don't use pkexec to start the reader-service"""

    def bootstrap_daemon():
        # The daemon gets fresh instances of everything, because as far as I remember
        # it runs in a separate process.
        global_config = GlobalConfig()
        global_uinputs = GlobalUInputs(UInput)
        mapping_parser = MappingParser(global_uinputs)

        return Daemon(
            global_config,
            global_uinputs,
            mapping_parser,
        )

    with (
        patch.object(
            os,
            "system",
            os_system_patch,
        ),
        patch.object(Daemon, "connect", bootstrap_daemon),
    ):
        yield


def clean_up_gui_test(test):
    logger.info("clean_up_gui_test")
    test.controller.stop_injecting()
    gtk_iteration()
    test.user_interface.on_gtk_close()
    test.user_interface.window.destroy()
    gtk_iteration()
    cleanup()

    # do this now, not when all tests are finished
    test.daemon.stop_all()
    if isinstance(test.daemon, Daemon):
        atexit.unregister(test.daemon.stop_all)


class GtkKeyEvent:
    def __init__(self, keyval):
        self.keyval = keyval

    def get_keyval(self):
        return True, self.keyval


@contextmanager
def patch_confirm_delete(
    user_interface: UserInterface,
    response=Gtk.ResponseType.ACCEPT,
):
    original_create_dialog = user_interface._create_dialog

    def _create_dialog_patch(*args, **kwargs):
        """A patch for the deletion confirmation that briefly shows the dialog."""
        confirm_cancel_dialog = original_create_dialog(*args, **kwargs)

        # the emitted signal causes the dialog to close
        GLib.timeout_add(
            100,
            lambda: confirm_cancel_dialog.emit("response", response),
        )

        # don't recursively call the patch
        Gtk.MessageDialog.run(confirm_cancel_dialog)

        confirm_cancel_dialog.run = lambda: response

        return confirm_cancel_dialog

    with patch.object(
        user_interface,
        "_create_dialog",
        _create_dialog_patch,
    ):
        # Tests are run during `yield`
        yield


class GuiTestBase(unittest.TestCase):
    def setUp(self):
        prepare_presets()
        with patch_services():
            (
                self.user_interface,
                self.controller,
                self.data_manager,
                self.message_broker,
                self.daemon,
                self.global_config,
            ) = launch()

        get = self.user_interface.get
        self.device_selection: Gtk.FlowBox = get("device_selection")
        self.preset_selection: Gtk.ComboBoxText = get("preset_selection")
        self.selection_label_listbox: Gtk.ListBox = get("selection_label_listbox")
        self.target_selection: Gtk.ComboBox = get("target-selector")
        self.recording_toggle: Gtk.ToggleButton = get("key_recording_toggle")
        self.recording_status: Gtk.ToggleButton = get("recording_status")
        self.status_bar: Gtk.Statusbar = get("status_bar")
        self.autoload_toggle: Gtk.Switch = get("preset_autoload_switch")
        self.code_editor: GtkSource.View = get("code_editor")
        self.output_box: GtkSource.View = get("output")

        self.delete_preset_btn: Gtk.Button = get("delete_preset")
        self.copy_preset_btn: Gtk.Button = get("copy_preset")
        self.create_preset_btn: Gtk.Button = get("create_preset")
        self.start_injector_btn: Gtk.Button = get("apply_preset")
        self.stop_injector_btn: Gtk.Button = get("stop_injection_preset_page")
        self.rename_btn: Gtk.Button = get("rename-button")
        self.rename_input: Gtk.Entry = get("preset_name_input")
        self.create_mapping_btn: Gtk.Button = get("create_mapping_button")
        self.delete_mapping_btn: Gtk.Button = get("delete-mapping")

        self._test_initial_state()

        self.grab_fails = False

        def grab(_):
            if self.grab_fails:
                raise OSError()

        evdev.InputDevice.grab = grab

        self.global_config._save_config()

        self.throttle(20)

        self.assertIsNotNone(self.data_manager.active_group)
        self.assertIsNotNone(self.data_manager.active_preset)

    def tearDown(self):
        clean_up_gui_test(self)

        # this is important, otherwise it keeps breaking things in the background
        self.assertIsNone(self.data_manager._reader_client._read_timeout)

        self.throttle(20)

    def get_code_input(self):
        buffer = self.code_editor.get_buffer()
        return buffer.get_text(
            buffer.get_start_iter(),
            buffer.get_end_iter(),
            True,
        )

    def _test_initial_state(self):
        # make sure each test deals with the same initial state
        self.assertEqual(self.controller.data_manager, self.data_manager)
        self.assertEqual(self.data_manager.active_group.key, "Foo Device")
        # if the modification-date from `prepare_presets` is not destroyed, preset3
        # should be selected as the newest one
        self.assertEqual(self.data_manager.active_preset.name, "preset3")
        self.assertEqual(self.data_manager.active_mapping.target_uinput, "keyboard")
        self.assertEqual(self.target_selection.get_active_id(), "keyboard")
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination([InputConfig(type=1, code=5)]),
        )
        self.assertEqual(
            self.data_manager.active_input_config, InputConfig(type=1, code=5)
        )
        self.assertGreater(
            len(self.user_interface.autocompletion._target_key_capabilities), 0
        )

    def _callTestMethod(self, method):
        """Retry all tests if they fail.

        GUI tests suddenly started to lag a lot and fail randomly, and even
        though that improved drastically, sometimes they still do.
        """
        attempts = 0
        while True:
            attempts += 1
            try:
                method()
                break
            except Exception as e:
                if attempts == 2:
                    raise e

            # try again
            print("Test failed, trying again...")
            self.tearDown()
            self.setUp()

    def throttle(self, time_=10):
        """Give GTK some time in ms to process everything."""
        # tests suddenly started to freeze my computer up completely and tests started
        # to fail. By using this (and by optimizing some redundant calls in the gui) it
        # worked again. EDIT: Might have been caused by my broken/bloated ssd. I'll
        # keep it in some places, since it did make the tests more reliable after all.
        for _ in range(time_ // 2):
            gtk_iteration()
            time.sleep(0.002)

    def set_focus(self, widget):
        logger.info("Focusing %s", widget)

        self.user_interface.window.set_focus(widget)

        self.throttle(20)

    def focus_source_view(self):
        # despite the focus and gtk_iterations, gtk never runs the event handlers for
        # the focus-in-event (_update_placeholder), which would clear the placeholder
        # text. Remove it manually, it can't be helped. Fun fact: when the
        # window gets destroyed, gtk runs the handler 10 times for good measure.
        # Lost one hour of my life on GTK again. It's gone! Forever! Use qt next time.
        source_view = self.code_editor
        self.set_focus(source_view)
        self.code_editor.get_buffer().set_text("")
        return source_view

    def get_selection_labels(self) -> List[MappingSelectionLabel]:
        return self.selection_label_listbox.get_children()

    def get_status_text(self):
        status_bar = self.user_interface.get("status_bar")
        return status_bar.get_message_area().get_children()[0].get_label()

    def get_unfiltered_symbol_input_text(self):
        buffer = self.code_editor.get_buffer()
        return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)

    def select_mapping(self, i: int):
        """Select one of the mappings of a preset.

        Parameters
        ----------
        i
            if -1, will select the last row,
            0 will select the uppermost row.
            1 will select the second row, and so on
        """
        selection_label = self.get_selection_labels()[i]
        self.selection_label_listbox.select_row(selection_label)
        logger.info(
            'Selecting mapping %s "%s"',
            selection_label.combination,
            selection_label.name,
        )
        gtk_iteration()
        return selection_label

    def add_mapping(self, mapping: Optional[Mapping] = None):
        self.controller.create_mapping()
        self.controller.load_mapping(InputCombination.empty_combination())
        gtk_iteration()
        if mapping:
            self.controller.update_mapping(**mapping.dict(exclude_defaults=True))
            gtk_iteration()

    def sleep(self, num_events):
        for _ in range(num_events * 2):
            time.sleep(EVENT_READ_TIMEOUT)
            gtk_iteration()

        time.sleep(1 / 30)  # one window iteration

        gtk_iteration()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import unittest

import evdev
from evdev.ecodes import (
    KEY_A,
    ABS_X,
)

from inputremapper.exceptions import EventNotHandled, UinputNotAvailable
from inputremapper.injection.global_uinputs import (
    FrontendUInput,
    GlobalUInputs,
    UInput,
)
from inputremapper.input_event import InputEvent
from tests.lib.cleanup import cleanup
from tests.lib.test_setup import test_setup


@test_setup
class TestFrontendUinput(unittest.TestCase):
    def setUp(self) -> None:
        cleanup()

    def test_init(self):
        name = "foo"
        capabilities = {1: [1, 2, 3], 2: [4, 5, 6]}
        uinput_defaults = FrontendUInput()
        uinput_custom = FrontendUInput(name=name, events=capabilities)

        self.assertEqual(uinput_defaults.name, "py-evdev-uinput")
        self.assertIsNone(uinput_defaults.capabilities())

        self.assertEqual(uinput_custom.name, name)
        self.assertEqual(uinput_custom.capabilities(), capabilities)


@test_setup
class TestGlobalUInputs(unittest.TestCase):
    def setUp(self) -> None:
        cleanup()

    def test_iter(self):
        global_uinputs = GlobalUInputs(FrontendUInput)
        for uinput in global_uinputs:
            self.assertIsInstance(uinput, evdev.UInput)

    def test_write(self):
        """Test write and write failure

        implicitly tests get_uinput and UInput.can_emit
        """
        global_uinputs = GlobalUInputs(UInput)
        global_uinputs.prepare_all()

        ev_1 = InputEvent.key(KEY_A, 1)
        ev_2 = InputEvent.abs(ABS_X, 10)

        keyboard = global_uinputs.get_uinput("keyboard")

        global_uinputs.write(ev_1.event_tuple, "keyboard")
        self.assertEqual(keyboard.write_count, 1)

        with self.assertRaises(EventNotHandled):
            global_uinputs.write(ev_2.event_tuple, "keyboard")

        with self.assertRaises(UinputNotAvailable):
            global_uinputs.write(ev_1.event_tuple, "foo")

    def test_creates_frontend_uinputs(self):
        frontend_uinputs = GlobalUInputs(FrontendUInput)
        frontend_uinputs.prepare_all()
        uinput = frontend_uinputs.get_uinput("keyboard")
        self.assertIsInstance(uinput, FrontendUInput)

    def test_creates_backend_service_uinputs(self):
        frontend_uinputs = GlobalUInputs(UInput)
        frontend_uinputs.prepare_all()
        uinput = frontend_uinputs.get_uinput("keyboard")
        self.assertIsInstance(uinput, UInput)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Build input-remapper including its translations and system-files.

pip, in many cases, fails to install data files, which need to go into system paths,
and instead puts them (despite them being absolute paths) into
/usr/lib/python3/inputremapper/usr/share/...

python3 setup.py install is deprecated

meson fails to install the module into a path that can actually be imported, and
its python features require one to specify each individual file of the module.

So instead, input-remapper uses a custom python solution. Hopefulls this works well
enough to prevent all ModuleNotFoundErrors in the future.
"""

import shutil
import os
import sys
from enum import Enum

from install.check_dependencies import check_dependencies
from install.data_files import build_data_files
from install.module import build_input_remapper_module
from install.language import make_lang

import argparse


class Components(str, Enum):
    data_files = "data_files"
    python_module = "python_module"


def parse_args():
    parser = argparse.ArgumentParser(description="Tool to install input-remapper with.")
    parser.add_argument(
        "--root",
        type=str,
        help=(
            "Where to install input-remapper to. For example ./build to prepare a "
            "package archive or for debugging, or / to install it to the system."
        ),
        required=True,
    )
    parser.add_argument(
        "--components",
        type=str, nargs='+',
        help=(
            f'A list of components to install. Default: --components '
            f'{Components.python_module.value} {Components.data_files.value}'
        ),
        default=[Components.python_module, Components.data_files],
        metavar="COMPONENT"
    )
    args = parser.parse_args()
    return args


def ask(msg) -> bool:
    answer = ""
    while answer not in ["y", "n"]:
        answer = input(f"{msg} [y/n] ").lower()
    return answer == "y"


def print_headline(message: str) -> None:
    print(f"\033[7m{message}\033[0m")


def main() -> None:
    args = parse_args()

    if os.path.exists(args.root) and not args.root.startswith("/"):
        delete = ask(f"{args.root} already exists. Delete?")
        if delete:
            shutil.rmtree(args.root)
        else:
            sys.exit(3)

    for component in args.components:
        if component not in [Components.data_files, Components.python_module]:
            raise ValueError(f"Unknown component {component}")

    if Components.data_files in args.components:
        print_headline(f'Installing component "{Components.data_files.value}"')
        build_data_files(args.root)
        make_lang(args.root)

    if Components.python_module in args.components:
        print_headline(f'Installing component "{Components.python_module.value}"')
        check_dependencies()
        build_input_remapper_module(args.root)


if __name__ == "__main__":
    main()



#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import shutil
import unittest
from packaging import version

from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    ABS_HAT0X,
    ABS_X,
    ABS_Y,
    ABS_RX,
    ABS_RY,
    EV_REL,
    REL_X,
    REL_Y,
    REL_WHEEL_HI_RES,
    REL_HWHEEL_HI_RES,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import UIMapping
from inputremapper.configs.migrations import Migrations
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.logging.logger import VERSION
from inputremapper.user import UserUtils
from tests.lib.test_setup import test_setup
from tests.lib.tmp import tmp


@test_setup
class TestMigrations(unittest.TestCase):
    def setUp(self):
        # some extra care to ensure those tests are not destroying actual presets
        self.assertTrue(UserUtils.home.startswith("/tmp"))
        self.assertTrue(PathUtils.config_path().startswith("/tmp"))
        self.assertTrue(PathUtils.get_preset_path().startswith("/tmp"))
        self.assertTrue(PathUtils.get_preset_path("foo", "bar").startswith("/tmp"))
        self.assertTrue(PathUtils.get_config_path().startswith("/tmp"))
        self.assertTrue(PathUtils.get_config_path("foo").startswith("/tmp"))

        self.v1_dir = os.path.join(UserUtils.home, ".config", "input-remapper")
        self.beta_dir = os.path.join(
            UserUtils.home, ".config", "input-remapper", "beta_1.6.0-beta"
        )

        global_uinputs = GlobalUInputs(UInput)
        global_uinputs.prepare_all()
        self.migrations = Migrations(global_uinputs)

    def test_migrate_suffix(self):
        old = os.path.join(PathUtils.config_path(), "config")
        new = os.path.join(PathUtils.config_path(), "config.json")

        try:
            os.remove(new)
        except FileNotFoundError:
            pass

        PathUtils.touch(old)
        with open(old, "w") as f:
            f.write("{}")

        self.migrations.migrate()
        self.assertTrue(os.path.exists(new))
        self.assertFalse(os.path.exists(old))

    def test_rename_config(self):
        old = os.path.join(UserUtils.home, ".config", "key-mapper")
        new = PathUtils.config_path()

        # we are not destroying our actual config files with this test
        self.assertTrue(new.startswith(tmp), f'Expected "{new}" to start with "{tmp}"')

        try:
            shutil.rmtree(new)
        except FileNotFoundError:
            pass

        old_config_json = os.path.join(old, "config.json")
        PathUtils.touch(old_config_json)
        with open(old_config_json, "w") as f:
            f.write('{"foo":"bar"}')

        self.migrations.migrate()

        self.assertTrue(os.path.exists(new))
        self.assertFalse(os.path.exists(old))

        new_config_json = os.path.join(new, "config.json")
        with open(new_config_json, "r") as f:
            moved_config = json.loads(f.read())
            self.assertEqual(moved_config["foo"], "bar")

    def test_wont_migrate_suffix(self):
        old = os.path.join(PathUtils.config_path(), "config")
        new = os.path.join(PathUtils.config_path(), "config.json")

        PathUtils.touch(new)
        with open(new, "w") as f:
            f.write("{}")

        PathUtils.touch(old)
        with open(old, "w") as f:
            f.write("{}")

        self.migrations.migrate()
        self.assertTrue(os.path.exists(new))
        self.assertTrue(os.path.exists(old))

    def test_migrate_preset(self):
        if os.path.exists(PathUtils.config_path()):
            shutil.rmtree(PathUtils.config_path())

        p1 = os.path.join(PathUtils.config_path(), "foo1", "bar1.json")
        p2 = os.path.join(PathUtils.config_path(), "foo2", "bar2.json")
        PathUtils.touch(p1)
        PathUtils.touch(p2)

        with open(p1, "w") as f:
            f.write("{}")

        with open(p2, "w") as f:
            f.write("{}")

        self.migrations.migrate()

        self.assertFalse(
            os.path.exists(os.path.join(PathUtils.config_path(), "foo1", "bar1.json"))
        )
        self.assertFalse(
            os.path.exists(os.path.join(PathUtils.config_path(), "foo2", "bar2.json"))
        )

        self.assertTrue(
            os.path.exists(
                os.path.join(PathUtils.config_path(), "presets", "foo1", "bar1.json")
            ),
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(PathUtils.config_path(), "presets", "foo2", "bar2.json")
            ),
        )

    def test_wont_migrate_preset(self):
        if os.path.exists(PathUtils.config_path()):
            shutil.rmtree(PathUtils.config_path())

        p1 = os.path.join(PathUtils.config_path(), "foo1", "bar1.json")
        p2 = os.path.join(PathUtils.config_path(), "foo2", "bar2.json")
        PathUtils.touch(p1)
        PathUtils.touch(p2)

        with open(p1, "w") as f:
            f.write("{}")

        with open(p2, "w") as f:
            f.write("{}")

        # already migrated
        PathUtils.mkdir(os.path.join(PathUtils.config_path(), "presets"))

        self.migrations.migrate()

        self.assertTrue(
            os.path.exists(os.path.join(PathUtils.config_path(), "foo1", "bar1.json"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(PathUtils.config_path(), "foo2", "bar2.json"))
        )

        self.assertFalse(
            os.path.exists(
                os.path.join(PathUtils.config_path(), "presets", "foo1", "bar1.json")
            ),
        )
        self.assertFalse(
            os.path.exists(
                os.path.join(PathUtils.config_path(), "presets", "foo2", "bar2.json")
            ),
        )

    def test_migrate_mappings(self):
        """Test if mappings are migrated correctly

        mappings like
        {(type, code): symbol} or {(type, code, value): symbol} should migrate
        to {InputCombination: {target: target, symbol: symbol, ...}}
        """
        path = os.path.join(
            PathUtils.config_path(), "presets", "Foo Device", "test.json"
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            json.dump(
                {
                    "mapping": {
                        f"{EV_KEY},1": "a",
                        f"{EV_KEY}, 2, 1": "BTN_B",  # can be mapped to "gamepad"
                        f"{EV_KEY}, 3, 1": "BTN_1",  # can not be mapped
                        f"{EV_ABS},{ABS_HAT0X},-1": "b",
                        f"{EV_ABS},1,1+{EV_ABS},2,-1+{EV_ABS},3,1": "c",
                        f"{EV_KEY}, 4, 1": ("d", "keyboard"),
                        f"{EV_KEY}, 5, 1": ("e", "foo"),  # unknown target
                        f"{EV_KEY}, 6, 1": ("key(a, b)", "keyboard"),  # broken macro
                        # ignored because broken
                        "3,1,1,2": "e",
                        "3": "e",
                        ",,+3,1,2": "g",
                        "": "h",
                    }
                },
                file,
            )
        self.migrations.migrate()
        # use UIMapping to also load invalid mappings
        preset = Preset(PathUtils.get_preset_path("Foo Device", "test"), UIMapping)
        preset.load()

        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=1)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=1)]),
                target_uinput="keyboard",
                output_symbol="a",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=2)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=2)]),
                target_uinput="gamepad",
                output_symbol="BTN_B",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=3)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=3)]),
                target_uinput="keyboard",
                output_symbol="BTN_1\n# Broken mapping:\n# No target can handle all specified keycodes",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=4)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=4)]),
                target_uinput="keyboard",
                output_symbol="d",
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_HAT0X, analog_threshold=-1)]
                )
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_HAT0X, analog_threshold=-1)]
                ),
                target_uinput="keyboard",
                output_symbol="b",
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination(
                    InputCombination.from_tuples(
                        (EV_ABS, 1, 1), (EV_ABS, 2, -1), (EV_ABS, 3, 1)
                    )
                ),
            ),
            UIMapping(
                input_combination=InputCombination(
                    InputCombination.from_tuples(
                        (EV_ABS, 1, 1), (EV_ABS, 2, -1), (EV_ABS, 3, 1)
                    ),
                ),
                target_uinput="keyboard",
                output_symbol="c",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=5)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=5)]),
                target_uinput="foo",
                output_symbol="e",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=6)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=6)]),
                target_uinput="keyboard",
                output_symbol="key(a, b)",
            ),
        )

        self.assertEqual(8, len(preset))

    def test_migrate_otherwise(self):
        path = os.path.join(
            PathUtils.config_path(), "presets", "Foo Device", "test.json"
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            json.dump(
                {
                    "mapping": {
                        f"{EV_KEY},1,1": ("otherwise + otherwise", "keyboard"),
                        f"{EV_KEY},2,1": ("bar($otherwise)", "keyboard"),
                        f"{EV_KEY},3,1": ("foo(otherwise=qux)", "keyboard"),
                        f"{EV_KEY},4,1": ("qux(otherwise).bar(otherwise = 1)", "foo"),
                        f"{EV_KEY},5,1": ("foo(otherwise1=2qux)", "keyboard"),
                    }
                },
                file,
            )

        self.migrations.migrate()

        preset = Preset(PathUtils.get_preset_path("Foo Device", "test"), UIMapping)
        preset.load()

        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=1)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=1)]),
                target_uinput="keyboard",
                output_symbol="otherwise + otherwise",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=2)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=2)]),
                target_uinput="keyboard",
                output_symbol="bar($otherwise)",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=3)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=3)]),
                target_uinput="keyboard",
                output_symbol="foo(else=qux)",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=4)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=4)]),
                target_uinput="foo",
                output_symbol="qux(otherwise).bar(else=1)",
            ),
        )
        self.assertEqual(
            preset.get_mapping(InputCombination([InputConfig(type=EV_KEY, code=5)])),
            UIMapping(
                input_combination=InputCombination([InputConfig(type=EV_KEY, code=5)]),
                target_uinput="keyboard",
                output_symbol="foo(otherwise1=2qux)",
            ),
        )

    def test_add_version(self):
        path = os.path.join(PathUtils.config_path(), "config.json")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            file.write("{}")

        self.migrations.migrate()
        self.assertEqual(
            version.parse(VERSION),
            self.migrations.config_version(),
        )

    def test_update_version(self):
        path = os.path.join(PathUtils.config_path(), "config.json")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            json.dump({"version": "0.1.0"}, file)

        self.migrations.migrate()
        self.assertEqual(
            version.parse(VERSION),
            self.migrations.config_version(),
        )

    def test_config_version(self):
        path = os.path.join(PathUtils.config_path(), "config.json")
        with open(path, "w") as file:
            file.write("{}")

        self.assertEqual("0.0.0", self.migrations.config_version().public)

        try:
            os.remove(path)
        except FileNotFoundError:
            pass

        self.assertEqual("0.0.0", self.migrations.config_version().public)

    def test_migrate_left_and_right_purpose(self):
        path = os.path.join(
            PathUtils.config_path(), "presets", "Foo Device", "test.json"
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            json.dump(
                {
                    "gamepad": {
                        "joystick": {
                            "left_purpose": "mouse",
                            "right_purpose": "wheel",
                            "pointer_speed": 50,
                            "x_scroll_speed": 10,
                            "y_scroll_speed": 20,
                        }
                    }
                },
                file,
            )
        self.migrations.migrate()

        preset = Preset(PathUtils.get_preset_path("Foo Device", "test"), UIMapping)
        preset.load()
        # 2 mappings for mouse
        # 2 mappings for wheel
        self.assertEqual(len(preset), 4)
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_X)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_X)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_X,
                gain=50 / 100,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_Y)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_Y)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_Y,
                gain=50 / 100,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_RX)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_RX)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_HWHEEL_HI_RES,
                gain=10,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_RY)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_RY)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_WHEEL_HI_RES,
                gain=20,
            ),
        )

    def test_migrate_left_and_right_purpose2(self):
        # same as above, but left and right is swapped

        path = os.path.join(
            PathUtils.config_path(), "presets", "Foo Device", "test.json"
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as file:
            json.dump(
                {
                    "gamepad": {
                        "joystick": {
                            "right_purpose": "mouse",
                            "left_purpose": "wheel",
                            "pointer_speed": 50,
                            "x_scroll_speed": 10,
                            "y_scroll_speed": 20,
                        }
                    }
                },
                file,
            )
        self.migrations.migrate()

        preset = Preset(PathUtils.get_preset_path("Foo Device", "test"), UIMapping)
        preset.load()
        # 2 mappings for mouse
        # 2 mappings for wheel
        self.assertEqual(len(preset), 4)
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_RX)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_RX)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_X,
                gain=50 / 100,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_RY)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_RY)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_Y,
                gain=50 / 100,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_X)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_X)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_HWHEEL_HI_RES,
                gain=10,
            ),
        )
        self.assertEqual(
            preset.get_mapping(
                InputCombination([InputConfig(type=EV_ABS, code=ABS_Y)])
            ),
            UIMapping(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_Y)]
                ),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_WHEEL_HI_RES,
                gain=20,
            ),
        )

    def _create_v1_setup(self):
        """Create all files needed to mimic an outdated v1 configuration."""
        device_name = "device_name"

        PathUtils.mkdir(os.path.join(self.v1_dir, "presets", device_name))
        v1_config = {"autoload": {device_name: "foo"}, "version": "1.0"}
        with open(os.path.join(self.v1_dir, "config.json"), "w") as file:
            json.dump(v1_config, file)
        # insert something outdated that will be migrated, to ensure the files are
        # first copied and then migrated.
        with open(
            os.path.join(self.v1_dir, "presets", device_name, "foo.json"), "w"
        ) as file:
            json.dump({"mapping": {f"{EV_KEY},1": "a"}}, file)

    def _create_beta_setup(self):
        """Create all files needed to mimic a beta configuration."""
        device_name = "device_name"

        # same here, but a different contents to tell the difference
        PathUtils.mkdir(os.path.join(self.beta_dir, "presets", device_name))
        beta_config = {"autoload": {device_name: "bar"}, "version": "1.6"}
        with open(os.path.join(self.beta_dir, "config.json"), "w") as file:
            json.dump(beta_config, file)
        with open(
            os.path.join(self.beta_dir, "presets", device_name, "bar.json"), "w"
        ) as file:
            json.dump(
                [
                    {
                        "input_combination": [
                            {"type": EV_KEY, "code": 1},
                        ],
                        "target_uinput": "keyboard",
                        "output_symbol": "b",
                        "mapping_type": "key_macro",
                    }
                ],
                file,
            )

    def test_prioritize_v1_over_beta_configs(self):
        # if both v1 and beta presets and config exist, migrate v1
        PathUtils.remove(PathUtils.get_config_path())

        device_name = "device_name"
        self._create_v1_setup()
        self._create_beta_setup()

        self.assertFalse(os.path.exists(PathUtils.get_preset_path(device_name, "foo")))
        self.assertFalse(os.path.exists(PathUtils.get_config_path("config.json")))

        self.migrations.migrate()

        self.assertTrue(os.path.exists(PathUtils.get_preset_path(device_name, "foo")))
        self.assertTrue(os.path.exists(PathUtils.get_config_path("config.json")))
        self.assertFalse(os.path.exists(PathUtils.get_preset_path(device_name, "bar")))

        # expect all original files to still exist
        self.assertTrue(os.path.join(self.v1_dir, "config.json"))
        self.assertTrue(os.path.join(self.v1_dir, "presets", "foo.json"))
        self.assertTrue(os.path.join(self.beta_dir, "config.json"))
        self.assertTrue(os.path.join(self.beta_dir, "presets", "bar.json"))

        # v1 configs should be in the v2 dir now, and migrated
        with open(PathUtils.get_config_path("config.json"), "r") as f:
            config_json = json.load(f)
            self.assertDictEqual(
                config_json, {"autoload": {device_name: "foo"}, "version": VERSION}
            )
        with open(PathUtils.get_preset_path(device_name, "foo.json"), "r") as f:
            os.system(f'cat { PathUtils.get_preset_path(device_name, "foo.json") }')
            preset_foo_json = json.load(f)
            self.assertEqual(
                preset_foo_json,
                [
                    {
                        "input_combination": [
                            {"type": EV_KEY, "code": 1},
                        ],
                        "target_uinput": "keyboard",
                        "output_symbol": "a",
                        "mapping_type": "key_macro",
                    }
                ],
            )

    def test_copy_over_beta_configs(self):
        # same as test_prioritize_v1_over_beta_configs, but only create the beta
        # directory without any v1 presets.
        PathUtils.remove(PathUtils.get_config_path())

        device_name = "device_name"
        self._create_beta_setup()

        self.assertFalse(os.path.exists(PathUtils.get_preset_path(device_name, "bar")))
        self.assertFalse(os.path.exists(PathUtils.get_config_path("config.json")))

        self.migrations.migrate()

        self.assertTrue(os.path.exists(PathUtils.get_preset_path(device_name, "bar")))
        self.assertTrue(os.path.exists(PathUtils.get_config_path("config.json")))

        # expect all original files to still exist
        self.assertTrue(os.path.join(self.beta_dir, "config.json"))
        self.assertTrue(os.path.join(self.beta_dir, "presets", "bar.json"))

        # beta configs should be in the v2 dir now
        with open(PathUtils.get_config_path("config.json"), "r") as f:
            config_json = json.load(f)
            self.assertDictEqual(
                config_json, {"autoload": {device_name: "bar"}, "version": VERSION}
            )
        with open(PathUtils.get_preset_path(device_name, "bar.json"), "r") as f:
            os.system(f'cat { PathUtils.get_preset_path(device_name, "bar.json") }')
            preset_foo_json = json.load(f)
            self.assertEqual(
                preset_foo_json,
                [
                    {
                        "input_combination": [
                            {"type": EV_KEY, "code": 1},
                        ],
                        "target_uinput": "keyboard",
                        "output_symbol": "b",
                        "mapping_type": "key_macro",
                    }
                ],
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import tracemalloc

from tests.lib.cleanup import cleanup, quick_cleanup
from tests.lib.fixture_pipes import create_fixture_pipes, remove_fixture_pipes
from tests.lib.is_service_running import is_service_running
from tests.lib.logger import update_inputremapper_verbosity, logger
from tests.lib.patches import create_patches


def test_setup(cls):
    """A class decorator to
    - apply the patches to all tests
    - check if the deamon is already running
    - create pipes to send events to the reader service
    - reset stuff automatically
    """
    original_setUp = cls.setUp
    original_tearDown = cls.tearDown
    original_setUpClass = cls.setUpClass
    original_tearDownClass = cls.tearDownClass

    tracemalloc.start()
    os.environ["UNITTEST"] = "1"
    update_inputremapper_verbosity()

    patches = create_patches()

    def resetPatches():
        # In case some patches carry a state (I don't remember, idk), stop and start
        # them from scratch
        for patch in patches:
            patch.stop()

        for patch in patches:
            patch.start()

    def setUpClass():
        logger.info("setUpClass %s", cls)

        if is_service_running():
            # let tests control daemon existance
            raise Exception("Expected the service not to be running already.")

        create_fixture_pipes()

        # I don't know. Somehow tearDownClass is called before the test, so lets
        # make sure the patches are started already when the class is set up, so that
        # an unpatched `prepare_all` doesn't take ages to finish, and doesn't do funky
        # stuff with the real evdev.
        resetPatches()

        original_setUpClass()

    def tearDownClass():
        logger.info("tearDownClass %s", cls)
        original_tearDownClass()

        remove_fixture_pipes()

        # Do the more thorough cleanup only after all tests of classes, because it
        # slows tests down. If this is required after each test, call it in your
        # tearDown method.
        cleanup()

    def setUp(self):
        logger.info("setUp %s", cls)

        resetPatches()

        original_setUp(self)

    def tearDown(self):
        logger.info("tearDown %s", cls)

        original_tearDown(self)

        quick_cleanup()

        resetPatches()

    cls.setUp = setUp
    cls.tearDown = tearDown
    cls.setUpClass = setUpClass
    cls.tearDownClass = tearDownClass

    return cls



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.
"""Make the systems/environments mapping of keys and codes accessible."""

import json
import re
import subprocess
from typing import Optional, List, Iterable, Tuple

import evdev

from inputremapper.configs.paths import PathUtils
from inputremapper.logging.logger import logger
from inputremapper.utils import is_service

DISABLE_NAME = "disable"

DISABLE_CODE = -1

# xkb uses keycodes that are 8 higher than those from evdev
XKB_KEYCODE_OFFSET = 8

XMODMAP_FILENAME = "xmodmap.json"

LAZY_LOAD = None


class KeyboardLayout:
    """Stores information about all available keycodes."""

    _mapping: Optional[dict] = LAZY_LOAD
    _xmodmap: Optional[List[Tuple[str, str]]] = LAZY_LOAD
    _case_insensitive_mapping: Optional[dict] = LAZY_LOAD

    def __getattribute__(self, wanted: str):
        """To lazy load keyboard_layout info only when needed.

        For example, this helps to keep logs of input-remapper-control clear when it
        doesn't need it the information.
        """
        lazy_loaded_attributes = ["_mapping", "_xmodmap", "_case_insensitive_mapping"]
        for lazy_loaded_attribute in lazy_loaded_attributes:
            if wanted != lazy_loaded_attribute:
                continue

            if object.__getattribute__(self, lazy_loaded_attribute) is LAZY_LOAD:
                # initialize _mapping and such with an empty dict, for populate
                # to write into
                object.__setattr__(self, lazy_loaded_attribute, {})
                object.__getattribute__(self, "populate")()

        return object.__getattribute__(self, wanted)

    def list_names(self, codes: Optional[Iterable[int]] = None) -> List[str]:
        """Get all possible names in the mapping, optionally filtered by codes.

        Parameters
        ----------
        codes: list of event codes
        """
        if not codes:
            return self._mapping.keys()

        return [name for name, code in self._mapping.items() if code in codes]

    def correct_case(self, symbol: str):
        """Return the correct casing for a symbol."""
        if symbol in self._mapping:
            return symbol
        # only if not e.g. both "a" and "A" are in the mapping
        return self._case_insensitive_mapping.get(symbol.lower(), symbol)

    def _use_xmodmap_symbols(self):
        """Look up xmodmap -pke, write xmodmap.json, and get the symbols."""
        try:
            xmodmap = subprocess.check_output(
                ["xmodmap", "-pke"],
                stderr=subprocess.STDOUT,
            ).decode()
        except FileNotFoundError:
            logger.info("Optional `xmodmap` command not found. This is not critical.")
            return
        except subprocess.CalledProcessError as e:
            logger.error('Call to `xmodmap -pke` failed with "%s"', e)
            return

        self._xmodmap = re.findall(r"(\d+) = (.+)\n", xmodmap + "\n")
        xmodmap_dict = self._find_legit_mappings()
        if len(xmodmap_dict) == 0:
            logger.info("`xmodmap -pke` did not yield any symbol")
            return

        # Write this stuff into the input-remapper config directory, because
        # the systemd service won't know the user sessions xmodmap.
        path = PathUtils.get_config_path(XMODMAP_FILENAME)
        PathUtils.touch(path)
        with open(path, "w") as file:
            logger.debug('Writing "%s"', path)
            json.dump(xmodmap_dict, file, indent=4)

        for name, code in xmodmap_dict.items():
            self._set(name, code)

    def _use_linux_evdev_symbols(self):
        """Look up the evdev constant names and use them."""
        for name, ecode in evdev.ecodes.ecodes.items():
            if name.startswith("KEY") or name.startswith("BTN"):
                self._set(name, ecode)

    def populate(self):
        """Get a mapping of all available names to their keycodes."""
        logger.debug("Gathering available keycodes")
        self.clear()

        if not is_service():
            # xmodmap is only available from within the login session.
            # The service that runs via systemd can't use this.
            self._use_xmodmap_symbols()

        self._use_linux_evdev_symbols()

        self._set(DISABLE_NAME, DISABLE_CODE)

    def update(self, mapping: dict):
        """Update this with new keys.

        Parameters
        ----------
        mapping
            maps from name to code. Make sure your keys are lowercase.
        """
        len_before = len(self._mapping)
        for name, code in mapping.items():
            self._set(name, code)

        logger.debug(
            "Updated keycodes with %d new ones", len(self._mapping) - len_before
        )

    def _set(self, name: str, code: int):
        """Map name to code."""
        self._mapping[str(name)] = code
        self._case_insensitive_mapping[str(name).lower()] = name

    def get(self, name: str) -> Optional[int]:
        """Return the code mapped to the key."""
        # the correct casing should be shown when asking the keyboard_layout
        # for stuff. indexing case insensitive to support old presets.
        if name not in self._mapping:
            # only if not e.g. both "a" and "A" are in the mapping
            name = self._case_insensitive_mapping.get(str(name).lower())

        return self._mapping.get(name)

    def clear(self):
        """Remove all mapped keys. Only needed for tests."""
        keys = list(self._mapping.keys())
        for key in keys:
            del self._mapping[key]

    def get_name(self, code: int):
        """Get the first matching name for the code."""
        for entry in self._xmodmap:
            if int(entry[0]) - XKB_KEYCODE_OFFSET == code:
                return entry[1].split()[0]

        # Fall back to the linux constants
        # This is especially important for BTN_LEFT and such
        btn_name = evdev.ecodes.BTN.get(code, None)
        if btn_name is not None:
            if type(btn_name) in [list, tuple]:
                # python-evdev >= 1.8.0 uses tuples
                return btn_name[0]

            return btn_name

        key_name = evdev.ecodes.KEY.get(code, None)
        if key_name is not None:
            if type(key_name) in [list, tuple]:
                # python-evdev >= 1.8.0 uses tuples
                return key_name[0]

            return key_name

        return None

    def _find_legit_mappings(self) -> dict:
        """From the parsed xmodmap list find usable symbols and their codes."""
        xmodmap_dict = {}
        for keycode, names in self._xmodmap:
            # there might be multiple, like:
            # keycode  64 = Alt_L Meta_L Alt_L Meta_L
            # keycode 204 = NoSymbol Alt_L NoSymbol Alt_L
            # Alt_L should map to code 64. Writing code 204 only works
            # if a modifier is applied at the same time. So take the first
            # one.
            name = names.split()[0]
            xmodmap_dict[name] = int(keycode) - XKB_KEYCODE_OFFSET

        return xmodmap_dict


# TODO DI
# this mapping represents the xmodmap output, which stays constant
keyboard_layout = KeyboardLayout()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Logging setup for input-remapper."""

import logging
import time
from typing import cast

from inputremapper.logging.formatter import ColorfulFormatter

from inputremapper.installation_info import VERSION, COMMIT_HASH


start = time.time()

previous_key_debug_log = None
previous_write_debug_log = None


class Logger(logging.Logger):
    def debug_mapping_handler(self, mapping_handler):
        """Parse the structure of a mapping_handler and log it."""
        if not self.isEnabledFor(logging.DEBUG):
            return

        lines_and_indent = self._parse_mapping_handler(mapping_handler)
        for line in lines_and_indent:
            indent = "    "
            msg = indent * line[1] + line[0]
            self._log(logging.DEBUG, msg, args=None)

    def write(self, key, uinput):
        """Log that an event is being written

        Parameters
        ----------
        key
            anything that can be string formatted, but usually a tuple of
            (type, code, value) tuples
        """
        # pylint: disable=protected-access
        if not self.isEnabledFor(logging.DEBUG):
            return

        global previous_write_debug_log

        str_key = repr(key)
        str_key = str_key.replace(",)", ")")

        msg = f'Writing {str_key} to "{uinput.name}"'

        if msg == previous_write_debug_log:
            # avoid some super spam from EV_ABS events
            return

        previous_write_debug_log = msg

        self._log(logging.DEBUG, msg, args=None, stacklevel=2)

    def _parse_mapping_handler(self, mapping_handler):
        indent = 0
        lines_and_indent = []
        while True:
            if isinstance(mapping_handler, list):
                for sub_handler in mapping_handler:
                    sub_list = self._parse_mapping_handler(sub_handler)
                    for line in sub_list:
                        line[1] += indent
                    lines_and_indent.extend(sub_list)
                break

            lines_and_indent.append([repr(mapping_handler), indent])
            try:
                mapping_handler = mapping_handler.child
            except AttributeError:
                break

            indent += 1
        return lines_and_indent

    def is_debug(self) -> bool:
        """True, if the logger is currently in DEBUG mode."""
        return self.level <= logging.DEBUG

    def log_info(self, name: str = "input-remapper") -> None:
        """Log version and name to the console."""
        logger.info(
            "%s %s %s https://github.com/sezanzeb/input-remapper",
            name,
            VERSION,
            COMMIT_HASH,
        )

        if EVDEV_VERSION:
            logger.info("python-evdev %s", EVDEV_VERSION)

        if self.is_debug():
            logger.warning(
                "Debug level will log all your keystrokes! Do not post this "
                "output in the internet if you typed in sensitive or private "
                "information with your device!"
            )

    def update_verbosity(self, debug: bool) -> None:
        """Set the logging verbosity according to the settings object."""
        if debug:
            self.setLevel(logging.DEBUG)
        else:
            self.setLevel(logging.INFO)

        for handler in self.handlers:
            handler.setFormatter(ColorfulFormatter(debug))

    @classmethod
    def bootstrap_logger(cls):
        # https://github.com/python/typeshed/issues/1801
        logging.setLoggerClass(cls)
        logger = cast(cls, logging.getLogger("input-remapper"))

        handler = logging.StreamHandler()
        handler.setFormatter(ColorfulFormatter(False))
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        logging.getLogger("asyncio").setLevel(logging.WARNING)
        return logger


logger = Logger.bootstrap_logger()


EVDEV_VERSION = None
try:
    from importlib.metadata import version

    EVDEV_VERSION = version("evdev")
except Exception as error:
    logger.info("Could not figure out the evdev version")
    logger.debug(error)

# check if the version is something like 1.5.0-beta or 1.5.0-beta.5
IS_BETA = "beta" in VERSION



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from unittest.mock import patch


def spy(obj, name):
    """Convenient wrapper for patch.object(..., ..., wraps=...)."""
    return patch.object(obj, name, wraps=obj.__getattribute__(name))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import time
import unittest

import gi

from inputremapper.gui.autocompletion import (
    get_incomplete_parameter,
    get_incomplete_function_name,
)

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "4")
gi.require_version("GLib", "2.0")
from gi.repository import Gtk, Gdk

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.gui.utils import gtk_iteration

from tests.lib.test_setup import test_setup
from tests.system.gui.gui_test_base import GuiTestBase


@test_setup
class TestAutocompletion(GuiTestBase):
    def press_key(self, keyval):
        event = Gdk.EventKey()
        event.keyval = keyval
        self.user_interface.autocompletion.navigate(None, event)

    def get_suggestions(self, autocompletion):
        return [
            row.get_children()[0].get_text()
            for row in autocompletion.list_box.get_children()
        ]

    def test_get_incomplete_parameter(self):
        def test(text, expected):
            text_view = Gtk.TextView()
            Gtk.TextView.do_insert_at_cursor(text_view, text)
            text_iter = text_view.get_iter_at_location(0, 0)[1]
            text_iter.set_offset(len(text))
            self.assertEqual(get_incomplete_parameter(text_iter), expected)

        test("bar(foo", "foo")
        test("bar(a=foo", "foo")
        test("bar(qux, foo", "foo")
        test("foo", "foo")
        test("bar + foo", "foo")

    def test_get_incomplete_function_name(self):
        def test(text, expected):
            text_view = Gtk.TextView()
            Gtk.TextView.do_insert_at_cursor(text_view, text)
            text_iter = text_view.get_iter_at_location(0, 0)[1]
            text_iter.set_offset(len(text))
            self.assertEqual(get_incomplete_function_name(text_iter), expected)

        test("bar().foo", "foo")
        test("bar()\n.foo", "foo")
        test("bar().\nfoo", "foo")
        test("bar(\nfoo", "foo")
        test("bar(\nqux=foo", "foo")
        test("bar(KEY_A,\nfoo", "foo")
        test("foo", "foo")

    def test_autocomplete_names(self):
        autocompletion = self.user_interface.autocompletion

        def setup(text):
            self.set_focus(self.code_editor)
            self.code_editor.get_buffer().set_text("")
            Gtk.TextView.do_insert_at_cursor(self.code_editor, text)
            self.throttle(200)
            text_iter = self.code_editor.get_iter_at_location(0, 0)[1]
            text_iter.set_offset(len(text))

        setup("disa")
        self.assertNotIn("KEY_A", self.get_suggestions(autocompletion))
        self.assertIn("disable", self.get_suggestions(autocompletion))

        setup(" + _A")
        self.assertIn("KEY_A", self.get_suggestions(autocompletion))
        self.assertNotIn("disable", self.get_suggestions(autocompletion))

    def test_autocomplete_key(self):
        self.controller.update_mapping(output_symbol="")
        gtk_iteration()

        self.set_focus(self.code_editor)
        self.code_editor.get_buffer().set_text("")

        complete_key_name = "Test_Foo_Bar"

        keyboard_layout.clear()
        keyboard_layout._set(complete_key_name, 1)
        keyboard_layout._set("KEY_A", 30)  # we need this for the UIMapping to work

        # it can autocomplete a combination inbetween other things
        incomplete = "qux_1\n +  + qux_2"
        Gtk.TextView.do_insert_at_cursor(self.code_editor, incomplete)
        Gtk.TextView.do_move_cursor(
            self.code_editor,
            Gtk.MovementStep.VISUAL_POSITIONS,
            -8,
            False,
        )

        Gtk.TextView.do_insert_at_cursor(self.code_editor, "foo")
        self.throttle(200)
        gtk_iteration()

        autocompletion = self.user_interface.autocompletion
        self.assertTrue(autocompletion.visible)

        self.press_key(Gdk.KEY_Down)
        self.press_key(Gdk.KEY_Return)
        self.throttle(200)
        gtk_iteration()

        # the first suggestion should have been selected

        modified_symbol = self.get_code_input()
        self.assertEqual(modified_symbol, f"qux_1\n + {complete_key_name} + qux_2")

        # try again, but a whitespace completes the word and so no autocompletion
        # should be shown
        Gtk.TextView.do_insert_at_cursor(self.code_editor, " + foo ")

        time.sleep(0.11)
        gtk_iteration()

        self.assertFalse(autocompletion.visible)

    def test_autocomplete_function(self):
        self.controller.update_mapping(output_symbol="")
        gtk_iteration()

        source_view = self.focus_source_view()

        incomplete = "key(KEY_A).\nepea"
        Gtk.TextView.do_insert_at_cursor(source_view, incomplete)

        time.sleep(0.11)
        gtk_iteration()

        autocompletion = self.user_interface.autocompletion
        self.assertTrue(autocompletion.visible)

        self.press_key(Gdk.KEY_Down)
        self.press_key(Gdk.KEY_Return)

        # the first suggestion should have been selected
        modified_symbol = self.get_code_input()
        self.assertEqual(modified_symbol, "key(KEY_A).\nrepeat")

    def test_close_autocompletion(self):
        self.controller.update_mapping(output_symbol="")
        gtk_iteration()

        source_view = self.focus_source_view()

        Gtk.TextView.do_insert_at_cursor(source_view, "KEY_")

        time.sleep(0.11)
        gtk_iteration()

        autocompletion = self.user_interface.autocompletion
        self.assertTrue(autocompletion.visible)

        self.press_key(Gdk.KEY_Down)
        self.press_key(Gdk.KEY_Escape)

        self.assertFalse(autocompletion.visible)

        symbol = self.get_code_input()
        self.assertEqual(symbol, "KEY_")

    def test_writing_still_works(self):
        self.controller.update_mapping(output_symbol="")
        gtk_iteration()
        source_view = self.focus_source_view()

        Gtk.TextView.do_insert_at_cursor(source_view, "KEY_")

        autocompletion = self.user_interface.autocompletion

        time.sleep(0.11)
        gtk_iteration()
        self.assertTrue(autocompletion.visible)

        # writing still works while an entry is selected
        self.press_key(Gdk.KEY_Down)

        Gtk.TextView.do_insert_at_cursor(source_view, "A")

        time.sleep(0.11)
        gtk_iteration()
        self.assertTrue(autocompletion.visible)

        Gtk.TextView.do_insert_at_cursor(source_view, "1234foobar")

        time.sleep(0.11)
        gtk_iteration()
        # no key matches this completion, so it closes again
        self.assertFalse(autocompletion.visible)

    def test_cycling(self):
        self.controller.update_mapping(output_symbol="")
        gtk_iteration()
        source_view = self.focus_source_view()

        Gtk.TextView.do_insert_at_cursor(source_view, "KEY_")

        autocompletion = self.user_interface.autocompletion

        time.sleep(0.11)
        gtk_iteration()
        self.assertTrue(autocompletion.visible)

        self.assertEqual(
            autocompletion.scrolled_window.get_vadjustment().get_value(), 0
        )

        # cycle to the end of the list because there is no element higher than index 0
        self.press_key(Gdk.KEY_Up)
        self.assertGreater(
            autocompletion.scrolled_window.get_vadjustment().get_value(), 0
        )

        # go back to the start, because it can't go down further
        self.press_key(Gdk.KEY_Down)
        self.assertEqual(
            autocompletion.scrolled_window.get_vadjustment().get_value(), 0
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio

from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.task import Task


class IfTapTask(Task):
    """If a key was pressed quickly.

    macro key pressed -> if_tap starts -> key released -> then

    macro key pressed -> released (does other stuff in the meantime)
    -> if_tap starts -> pressed -> released -> then
    """

    argument_configs = [
        ArgumentConfig(
            name="then",
            position=0,
            types=[Macro, None],
            default=None,
        ),
        ArgumentConfig(
            name="else",
            position=1,
            types=[Macro, None],
            default=None,
        ),
        ArgumentConfig(
            name="timeout",
            position=2,
            types=[int, float],
            default=300,
        ),
    ]

    async def run(self, callback) -> None:
        then = self.get_argument("then").get_value()
        else_ = self.get_argument("else").get_value()
        timeout = self.get_argument("timeout").get_value() / 1000

        try:
            await asyncio.wait_for(self._wait(), timeout)
            if then:
                await then.run(callback)
        except asyncio.TimeoutError:
            if else_:
                await else_.run(callback)

    async def _wait(self):
        """Wait for a release, or if nothing pressed yet, a press and release."""
        if self.is_holding():
            await self._trigger_release_event.wait()
        else:
            await self._trigger_press_event.wait()
            await self._trigger_release_event.wait()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import time

import evdev
from evdev.ecodes import EV_REL

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.logging.logger import logger


class RelToBtnHandler(MappingHandler):
    """Handler which transforms an EV_REL to a button event
    and sends that to a sub_handler

    adheres to the MappingHandler protocol
    """

    _active: bool
    _input_config: InputConfig
    _last_activation: float
    _sub_handler: InputEventHandler

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        super().__init__(combination, mapping, global_uinputs)

        self._active = False
        self._input_config = combination[0]
        self._last_activation = time.time()
        self._abort_release = False
        assert self._input_config.analog_threshold != 0
        assert len(combination) == 1

    def __str__(self):
        return f'RelToBtnHandler for "{self._input_config}"'

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return self._sub_handler

    async def _stage_release(
        self,
        source: InputEvent,
        suppress: bool,
    ):
        while time.time() < self._last_activation + self.mapping.release_timeout:
            await asyncio.sleep(1 / self.mapping.rel_rate)

        if self._abort_release:
            self._abort_release = False
            return

        event = InputEvent(
            0,
            0,
            *self._input_config.type_and_code,
            value=0,
            actions=(EventActions.as_key,),
            origin_hash=self._input_config.origin_hash,
        )
        logger.debug("Sending %s to sub_handler", event)
        self._sub_handler.notify(event, source, suppress)
        self._active = False

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        assert event.type == EV_REL
        if event.input_match_hash != self._input_config.input_match_hash:
            return False

        assert (threshold := self._input_config.analog_threshold)
        value = event.value
        if (value < threshold > 0) or (value > threshold < 0):
            if self._active:
                # the axis is below the threshold and the stage_release
                # function is running
                if self.mapping.force_release_timeout:
                    # consume the event
                    return True
                event = event.modify(value=0, actions=(EventActions.as_key,))
                logger.debug("Sending %s to sub_handler", event)
                self._abort_release = True
            else:
                # don't consume the event.
                # We could return True to consume events
                return False
        else:
            # the axis is above the threshold
            if not self._active:
                asyncio.ensure_future(self._stage_release(source, suppress))
            if value >= threshold > 0:
                direction = EventActions.positive_trigger
            else:
                direction = EventActions.negative_trigger
            self._last_activation = time.time()
            event = event.modify(value=1, actions=(EventActions.as_key, direction))

        self._active = bool(event.value)
        # logger.debug("Sending %s to sub_handler", event)
        return self._sub_handler.notify(event, source=source, suppress=suppress)

    def reset(self) -> None:
        if self._active:
            self._abort_release = True

        self._active = False
        self._sub_handler.reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestHoldKeys(MacroTestBase):
    async def test_hold_keys(self):
        macro = Parser.parse(
            "set(foo, b).hold_keys(a, $foo, c)", self.context, DummyMapping
        )
        # press first
        macro.press_trigger()
        # then run, just like how it is going to happen during runtime
        asyncio.ensure_future(macro.run(self.handler))

        code_a = keyboard_layout.get("a")
        code_b = keyboard_layout.get("b")
        code_c = keyboard_layout.get("c")

        await asyncio.sleep(0.2)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, code_a, 1),
                (EV_KEY, code_b, 1),
                (EV_KEY, code_c, 1),
            ],
        )

        macro.release_trigger()

        await asyncio.sleep(0.2)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, code_a, 1),
                (EV_KEY, code_b, 1),
                (EV_KEY, code_c, 1),
                (EV_KEY, code_c, 0),
                (EV_KEY, code_b, 0),
                (EV_KEY, code_a, 0),
            ],
        )

    async def test_hold_keys_broken(self):
        # Won't run any of the keys when one of them is invalid
        macro = Parser.parse(
            "set(foo, broken).hold_keys(a, $foo, c)", self.context, DummyMapping
        )
        # press first
        macro.press_trigger()
        # then run, just like how it is going to happen during runtime
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.2)
        self.assertListEqual(self.result, [])
        macro.release_trigger()
        await asyncio.sleep(0.2)
        self.assertListEqual(self.result, [])

    async def test_aldjfakl(self):
        repeats = 5

        macro = Parser.parse(
            f"repeat({repeats}, key(k))",
            self.context,
            DummyMapping,
        )

        self.assertEqual(self.count_child_macros(macro), 1)

    async def test_run_plus_syntax(self):
        macro = Parser.parse("a + b + c + d", self.context, DummyMapping)

        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.2)
        self.assertTrue(macro.tasks[0].is_holding())

        # starting from the left, presses each one down
        self.assertEqual(self.result[0], (EV_KEY, keyboard_layout.get("a"), 1))
        self.assertEqual(self.result[1], (EV_KEY, keyboard_layout.get("b"), 1))
        self.assertEqual(self.result[2], (EV_KEY, keyboard_layout.get("c"), 1))
        self.assertEqual(self.result[3], (EV_KEY, keyboard_layout.get("d"), 1))

        # and then releases starting with the previously pressed key
        macro.release_trigger()
        await asyncio.sleep(0.2)
        self.assertFalse(macro.tasks[0].is_holding())
        self.assertEqual(self.result[4], (EV_KEY, keyboard_layout.get("d"), 0))
        self.assertEqual(self.result[5], (EV_KEY, keyboard_layout.get("c"), 0))
        self.assertEqual(self.result[6], (EV_KEY, keyboard_layout.get("b"), 0))
        self.assertEqual(self.result[7], (EV_KEY, keyboard_layout.get("a"), 0))

    async def test_raises_error(self):
        self.assertRaises(
            MacroError, Parser.parse, "hold_keys(a, broken, b)", self.context
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.
import time
import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestRepeat(MacroTestBase):
    async def test_1(self):
        start = time.time()
        repeats = 20

        macro = Parser.parse(
            f"repeat({repeats}, key(k)).repeat(1, key(k))",
            self.context,
            DummyMapping,
        )
        k_code = keyboard_layout.get("k")

        await macro.run(self.handler)
        keystroke_sleep = DummyMapping.macro_key_sleep_ms
        sleep_time = 2 * repeats * keystroke_sleep / 1000
        self.assertGreater(time.time() - start, sleep_time * 0.9)

        # This test is rather lax, timing in github actions is slow
        self.assertLess(time.time() - start, sleep_time * 1.4)

        self.assertListEqual(
            self.result,
            [(EV_KEY, k_code, 1), (EV_KEY, k_code, 0)] * (repeats + 1),
        )

        self.assertEqual(self.count_child_macros(macro), 2)

        self.assertEqual(len(macro.tasks[0].child_macros), 1)
        self.assertEqual(len(macro.tasks[0].child_macros[0].tasks), 1)
        self.assertEqual(len(macro.tasks[0].child_macros[0].tasks[0].child_macros), 0)

        self.assertEqual(len(macro.tasks[1].child_macros), 1)
        self.assertEqual(len(macro.tasks[1].child_macros[0].tasks), 1)
        self.assertEqual(len(macro.tasks[1].child_macros[0].tasks[0].child_macros), 0)

    async def test_2(self):
        start = time.time()
        macro = Parser.parse("repeat(3, key(m).w(100))", self.context, DummyMapping)
        m_code = keyboard_layout.get("m")
        await macro.run(self.handler)

        keystroke_time = 6 * DummyMapping.macro_key_sleep_ms
        total_time = keystroke_time + 300
        total_time /= 1000

        self.assertGreater(time.time() - start, total_time * 0.9)
        self.assertLess(time.time() - start, total_time * 1.2)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, m_code, 1),
                (EV_KEY, m_code, 0),
                (EV_KEY, m_code, 1),
                (EV_KEY, m_code, 0),
                (EV_KEY, m_code, 1),
                (EV_KEY, m_code, 0),
            ],
        )
        self.assertEqual(self.count_child_macros(macro), 1)
        self.assertEqual(len(macro.tasks), 1)
        self.assertEqual(len(macro.tasks[0].child_macros), 1)
        self.assertEqual(len(macro.tasks[0].child_macros[0].tasks), 2)
        self.assertEqual(len(macro.tasks[0].child_macros[0].tasks[0].child_macros), 0)
        self.assertEqual(len(macro.tasks[0].child_macros[0].tasks[1].child_macros), 0)

    async def test_not_an_int(self):
        # the first parameter for `repeat` requires an integer, not "foo",
        # which makes `repeat` throw
        macro = Parser.parse(
            'set(a, "foo").repeat($a, key(KEY_A)).key(KEY_B)',
            self.context,
            DummyMapping,
        )

        try:
            await macro.run(self.handler)
        except MacroError as e:
            self.assertIn("foo", str(e))

        self.assertFalse(macro.running)

        # key(KEY_B) is not executed, the macro stops
        self.assertListEqual(self.result, [])

    async def test_raises_error(self):
        self.assertRaises(MacroError, Parser.parse, "r(1)", self.context)
        self.assertRaises(MacroError, Parser.parse, "repeat(a, k(1))", self.context)
        Parser.parse("repeat($a, k(1))", self.context)  # no error
        Parser.parse("repeat(2, k(1))", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, 'repeat("2", k(1))', self.context)
        self.assertRaises(MacroError, Parser.parse, "r(1, 1)", self.context)
        self.assertRaises(MacroError, Parser.parse, "r(k(1), 1)", self.context)
        Parser.parse("r(1, macro=k(1))", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "r(a=1, b=k(1))", self.context)
        self.assertRaises(
            MacroError,
            Parser.parse,
            "r(repeats=1, macro=k(1), a=2)",
            self.context,
        )
        self.assertRaises(
            MacroError,
            Parser.parse,
            "r(repeats=1, macro=k(1), repeats=2)",
            self.context,
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import unittest

import evdev

from inputremapper.configs.paths import PathUtils
from inputremapper.groups import (
    _FindGroups,
    groups,
    classify,
    DeviceType,
    _Group,
)
from tests.lib.fixtures import fixtures, keyboard_keys
from tests.lib.test_setup import test_setup


class FakePipe:
    groups = None

    def send(self, groups):
        self.groups = groups


@test_setup
class TestGroups(unittest.TestCase):
    def test_group(self):
        group = _Group(
            paths=["/dev/a", "/dev/b", "/dev/c"],
            names=["name_bar", "name_a", "name_foo"],
            types=[DeviceType.MOUSE, DeviceType.KEYBOARD, DeviceType.UNKNOWN],
            key="key",
        )
        self.assertEqual(group.name, "name_a")
        self.assertEqual(group.key, "key")
        self.assertEqual(
            group.get_preset_path("preset1234"),
            os.path.join(
                PathUtils.config_path(),
                "presets",
                group.name,
                "preset1234.json",
            ),
        )

    def test_find_groups(self):
        pipe = FakePipe()
        _FindGroups(pipe).run()
        self.assertIsInstance(pipe.groups, str)

        groups.loads(pipe.groups)
        self.maxDiff = None
        self.assertEqual(
            groups.dumps(),
            json.dumps(
                [
                    json.dumps(
                        {
                            "paths": [
                                "/dev/input/event1",
                            ],
                            "names": ["Foo Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Foo Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": [
                                "/dev/input/event11",
                                "/dev/input/event10",
                                "/dev/input/event13",
                                "/dev/input/event15",
                            ],
                            "names": [
                                "Foo Device foo",
                                "Foo Device",
                                "Foo Device",
                                "Foo Device bar",
                            ],
                            "types": [
                                DeviceType.GAMEPAD,
                                DeviceType.KEYBOARD,
                                DeviceType.MOUSE,
                            ],
                            "key": "Foo Device 2",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event20"],
                            "names": ["Bar Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Bar Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event30"],
                            "names": ["gamepad"],
                            "types": [DeviceType.GAMEPAD],
                            "key": "gamepad",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event40"],
                            "names": ["input-remapper Bar Device"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "input-remapper Bar Device",
                        }
                    ),
                    json.dumps(
                        {
                            "paths": ["/dev/input/event52"],
                            "names": ["Qux/[Device]?"],
                            "types": [DeviceType.KEYBOARD],
                            "key": "Qux/[Device]?",
                        }
                    ),
                ]
            ),
        )

        groups2 = json.dumps(
            [group.dumps() for group in groups.filter(include_inputremapper=True)]
        )
        self.assertEqual(pipe.groups, groups2)

    def test_list_group_names(self):
        self.assertListEqual(
            groups.list_group_names(),
            [
                "Foo Device",
                "Foo Device",
                "Bar Device",
                "gamepad",
                "Qux/[Device]?",
            ],
        )

    def test_filter(self):
        # by default no input-remapper devices are present
        filtered = groups.filter()
        keys = [group.key for group in filtered]
        self.assertIn("Foo Device 2", keys)
        self.assertNotIn("input-remapper Bar Device", keys)

    def test_skip_camera(self):
        fixtures["/foo/bar"] = {
            "name": "camera",
            "phys": "abcd1",
            "info": evdev.DeviceInfo(1, 2, 3, 4),
            "capabilities": {evdev.ecodes.EV_KEY: [evdev.ecodes.KEY_CAMERA]},
        }

        groups.refresh()
        self.assertIsNone(groups.find(name="camera"))
        self.assertIsNotNone(groups.find(name="gamepad"))

    def test_device_with_only_ev_abs(self):
        # As Input Mapper can now map axes to buttons,
        # a single EV_ABS device is valid for mapping.
        fixtures["/foo/bar"] = {
            "name": "qux",
            "phys": "abcd2",
            "info": evdev.DeviceInfo(1, 2, 3, 4),
            "capabilities": {evdev.ecodes.EV_ABS: [evdev.ecodes.ABS_X]},
        }

        groups.refresh()
        self.assertIsNotNone(groups.find(name="gamepad"))
        self.assertIsNotNone(groups.find(name="qux"))

    def test_device_with_no_capabilities(self):
        fixtures["/foo/bar"] = {
            "name": "nulcap",
            "phys": "abcd3",
            "info": evdev.DeviceInfo(1, 2, 3, 4),
            "capabilities": {},
        }

        groups.refresh()
        self.assertIsNotNone(groups.find(name="gamepad"))
        self.assertIsNone(groups.find(name="nulcap"))

    def test_duplicate_device(self):
        fixtures["/dev/input/event100"] = {
            "capabilities": {evdev.ecodes.EV_KEY: keyboard_keys},
            "phys": "usb-0000:03:00.0-3/input1",
            "info": evdev.device.DeviceInfo(2, 1, 2, 1),
            "name": "Foo Device",
        }
        groups.refresh()

        group1 = groups.find(key="Foo Device")
        group2 = groups.find(key="Foo Device 2")
        group3 = groups.find(key="Foo Device 3")

        self.assertIn("/dev/input/event1", group1.paths)
        self.assertIn("/dev/input/event10", group2.paths)
        self.assertIn("/dev/input/event100", group3.paths)

        self.assertEqual(group1.key, "Foo Device")
        self.assertEqual(group2.key, "Foo Device 2")
        self.assertEqual(group3.key, "Foo Device 3")

        self.assertEqual(group1.name, "Foo Device")
        self.assertEqual(group2.name, "Foo Device")
        self.assertEqual(group3.name, "Foo Device")

    def test_classify(self):
        # properly detects if the device is a gamepad
        EV_ABS = evdev.ecodes.EV_ABS
        EV_KEY = evdev.ecodes.EV_KEY
        EV_REL = evdev.ecodes.EV_REL

        class FakeDevice:
            def __init__(self, capabilities):
                self.c = capabilities

            def capabilities(self, absinfo):
                assert not absinfo
                return self.c

        """Gamepads"""

        self.assertEqual(
            classify(
                FakeDevice(
                    {
                        EV_ABS: [evdev.ecodes.ABS_X, evdev.ecodes.ABS_Y],
                        EV_KEY: [evdev.ecodes.BTN_A],
                    }
                )
            ),
            DeviceType.GAMEPAD,
        )

        """Mice"""

        self.assertEqual(
            classify(
                FakeDevice(
                    {
                        EV_REL: [
                            evdev.ecodes.REL_X,
                            evdev.ecodes.REL_Y,
                            evdev.ecodes.REL_WHEEL,
                        ],
                        EV_KEY: [evdev.ecodes.BTN_LEFT],
                    }
                )
            ),
            DeviceType.MOUSE,
        )

        """Keyboard"""

        self.assertEqual(
            classify(FakeDevice({EV_KEY: [evdev.ecodes.KEY_A]})), DeviceType.KEYBOARD
        )

        """Touchpads"""

        self.assertEqual(
            classify(
                FakeDevice(
                    {
                        EV_KEY: [evdev.ecodes.KEY_A],
                        EV_ABS: [evdev.ecodes.ABS_MT_POSITION_X],
                    }
                )
            ),
            DeviceType.TOUCHPAD,
        )

        """Graphics tablets"""

        self.assertEqual(
            classify(
                FakeDevice(
                    {
                        EV_ABS: [evdev.ecodes.ABS_X, evdev.ecodes.ABS_Y],
                        EV_KEY: [evdev.ecodes.BTN_STYLUS],
                    }
                )
            ),
            DeviceType.GRAPHICS_TABLET,
        )

        """Weird combos"""

        self.assertEqual(
            classify(
                FakeDevice(
                    {
                        EV_ABS: [evdev.ecodes.ABS_X, evdev.ecodes.ABS_Y],
                        EV_KEY: [evdev.ecodes.KEY_1],
                    }
                )
            ),
            DeviceType.UNKNOWN,
        )

        self.assertEqual(
            classify(
                FakeDevice({EV_ABS: [evdev.ecodes.ABS_X], EV_KEY: [evdev.ecodes.BTN_A]})
            ),
            DeviceType.UNKNOWN,
        )

        self.assertEqual(
            classify(FakeDevice({EV_KEY: [evdev.ecodes.BTN_A]})), DeviceType.UNKNOWN
        )

        self.assertEqual(
            classify(FakeDevice({EV_ABS: [evdev.ecodes.ABS_X]})), DeviceType.UNKNOWN
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
from itertools import chain
from typing import List, Dict, TYPE_CHECKING, Optional, Tuple, Union

from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.argument import (
    Argument,
    ArgumentConfig,
    ArgumentFlags,
)
from inputremapper.injection.macros.macro import Macro, InjectEventCallback
from inputremapper.logging.logger import logger

if TYPE_CHECKING:
    from inputremapper.injection.mapping_handlers.mapping_handler import EventListener
    from inputremapper.injection.macros.raw_value import RawValue
    from inputremapper.injection.context import Context
    from inputremapper.configs.mapping import Mapping


class Task:
    """Base Class for functions like `if_eq` or `key`

    A macro like `key(a).key(b)` will contain two instances of this class.
    """

    argument_configs: List[ArgumentConfig]
    arguments: Dict[str, Argument]
    mapping: Mapping

    # The context is None during frontend-parsing/validation I believe
    context: Optional[Context]

    child_macros: List[Macro]

    def __init__(
        self,
        positional_args: List[RawValue],
        keyword_args: Dict[str, RawValue],
        context: Optional[Context],
        mapping: Mapping,
    ) -> None:
        self.context = context
        self.mapping = mapping
        self.child_macros = []

        self._validate_argument_configs()

        self.arguments = {
            argument_config.name: Argument(argument_config, mapping)
            for argument_config in self.argument_configs
        }

        self._setup_asyncio_events()

        for argument in self.arguments.values():
            self._initialize_argument(argument, keyword_args, positional_args)

        self._initialize_spread_arg(positional_args)

        for raw_value in chain(keyword_args.values(), positional_args):
            if isinstance(raw_value.value, Macro):
                self.child_macros.append(raw_value.value)

    async def run(self, callback: InjectEventCallback) -> None:
        """Macro logic goes here.

        Call the callback with the type, code and value that should be injected.
        """
        raise NotImplementedError()

    def add_event_listener(self, listener: EventListener) -> None:
        """Listeners get each event from the source device.

        After all listeners are done, the event will go into the mapping handlers.

        Make sure to remove your event_listener once you are done.
        """
        # The context will be there when the macro is parsed by the service
        assert self.context is not None
        self.context.listeners.add(listener)

    def remove_event_listener(self, listener: EventListener) -> None:
        assert self.context is not None
        self.context.listeners.remove(listener)

    @classmethod
    def get_macro_argument_names(cls):
        return [argument_config.name for argument_config in cls.argument_configs]

    @classmethod
    def get_num_parameters(cls) -> Tuple[int, Union[int, float]]:
        """Get the number of required parameters and the maximum number of parameters."""
        min_num_args = 0
        argument_configs = cls.argument_configs
        max_num_args: Union[int, float] = len(argument_configs)
        for argument_config in argument_configs:
            if argument_config.position == ArgumentFlags.spread:
                # 0 or more
                max_num_args = float("inf")
                continue

            if argument_config.is_required():
                min_num_args += 1

        return min_num_args, max_num_args

    def get_argument(self, argument_name) -> Argument:
        return self.arguments[argument_name]

    def press_trigger(self) -> None:
        """The user pressed the trigger key down."""
        for macro in self.child_macros:
            macro.press_trigger()

        if self.is_holding():
            logger.error("Already holding")
            return

        self._trigger_release_event.clear()
        self._trigger_press_event.set()

    def release_trigger(self) -> None:
        """The user released the trigger key."""
        if not self.is_holding():
            return

        self._trigger_release_event.set()
        self._trigger_press_event.clear()

        for macro in self.child_macros:
            macro.release_trigger()

    def is_holding(self) -> bool:
        """Check if the macro is waiting for a key to be released."""
        return not self._trigger_release_event.is_set()

    async def keycode_pause(self, _=None) -> None:
        """To add a pause between keystrokes.

        This was needed at some point because it appeared that injecting keys too
        fast will prevent them from working. It probably depends on the environment.
        """
        await asyncio.sleep(self.mapping.macro_key_sleep_ms / 1000)

    def _initialize_spread_arg(
        self,
        positional_args: List[RawValue],
    ) -> None:
        """Put all positional arguments that aren't used into the spread argument."""
        spread_argument: Optional[Argument] = None
        for argument in self.arguments.values():
            if argument.position == ArgumentFlags.spread:
                spread_argument = argument
                break

        if spread_argument is None:
            return

        remaining_positional_args = [*positional_args]

        for argument in self.arguments.values():
            if argument.position != ArgumentFlags.spread and argument.position < len(
                remaining_positional_args
            ):
                del remaining_positional_args[argument.position]

        spread_argument.initialize_variables(remaining_positional_args)

    def _find_argument_by_position(self, position: int) -> Optional[Argument]:
        for argument in self.arguments.values():
            if argument.position == position:
                return argument

        return None

    def _setup_asyncio_events(self) -> None:
        # Can be used to wait for the press and release of the input event/key, that is
        # configured as the trigger of the macro, via asyncio.
        self._trigger_release_event = asyncio.Event()
        self._trigger_press_event = asyncio.Event()
        # released by default
        self._trigger_release_event.set()
        self._trigger_press_event.clear()

    def _initialize_argument(
        self,
        argument: Argument,
        keyword_args: Dict[str, RawValue],
        positional_args: List[RawValue],
    ) -> None:
        if argument.position == ArgumentFlags.spread:
            # Will get all the remaining positional arguments afterward.
            return

        for name, value in keyword_args.items():
            if argument.name == name:
                argument.initialize_variable(value)
                return

        if argument.position < len(positional_args):
            argument.initialize_variable(positional_args[argument.position])
            return

        if not argument.is_required():
            argument.initialize_default()
            return

        # This shouldn't be possible, the parser should have ensured things are valid
        # already.
        raise MacroError(f"Could not initialize argument {argument.name}")

    def _validate_argument_configs(self):
        # Might help during development
        positions = set()
        names = set()
        for argument_config in self.argument_configs:
            position = argument_config.position
            if position in positions:
                raise MacroError(f"Duplicate position {positions} in ArgumentConfig")
            positions.add(position)

            name = argument_config.name
            if name in names:
                raise MacroError(f"Duplicate name {name} in ArgumentConfig")
            names.add(name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import unittest

from evdev.ecodes import (
    EV_KEY,
    ABS_Y,
)

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from inputremapper.input_event import InputEvent
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestIfSingle(MacroTestBase):
    async def test_if_single(self):
        macro = Parser.parse("if_single(key(x), key(y))", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 2)

        a = keyboard_layout.get("a")
        x = keyboard_layout.get("x")

        await self.trigger_sequence(macro, InputEvent.key(a, 1))
        await asyncio.sleep(0.1)
        await self.release_sequence(macro, InputEvent.key(a, 0))
        # the key that triggered the macro is released
        await asyncio.sleep(0.1)

        self.assertListEqual(self.result, [(EV_KEY, x, 1), (EV_KEY, x, 0)])
        self.assertFalse(macro.running)

    async def test_if_single_ignores_releases(self):
        # the timeout won't break the macro, everything happens well within that
        # timeframe.
        macro = Parser.parse(
            "if_single(key(x), else=key(y), timeout=100000)",
            self.context,
            DummyMapping,
        )
        self.assertEqual(self.count_child_macros(macro), 2)

        a = keyboard_layout.get("a")
        b = keyboard_layout.get("b")

        x = keyboard_layout.get("x")
        y = keyboard_layout.get("y")

        # pressing the macro key
        await self.trigger_sequence(macro, InputEvent.key(a, 1))
        await asyncio.sleep(0.05)

        # if_single only looks out for newly pressed keys,
        # it doesn't care if keys were released that have been
        # pressed before if_single. This was decided because it is a lot
        # less tricky and more fluently to use if you type fast
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(InputEvent.key(b, 0)))
        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [])

        # releasing the actual key triggers if_single
        await asyncio.sleep(0.05)
        await self.release_sequence(macro, InputEvent.key(a, 0))
        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [(EV_KEY, x, 1), (EV_KEY, x, 0)])
        self.assertFalse(macro.running)

    async def test_if_not_single(self):
        # Will run the `else` macro if another key is pressed.
        # Also works if if_single is a child macro, i.e. the event is passed to it
        # from the outside macro correctly.
        macro = Parser.parse(
            "repeat(1, if_single(then=key(x), else=key(y)))",
            self.context,
            DummyMapping,
        )
        self.assertEqual(self.count_child_macros(macro), 3)
        self.assertEqual(self.count_tasks(macro), 4)

        a = keyboard_layout.get("a")
        b = keyboard_layout.get("b")

        x = keyboard_layout.get("x")
        y = keyboard_layout.get("y")

        # press the trigger key
        await self.trigger_sequence(macro, InputEvent.key(a, 1))
        await asyncio.sleep(0.1)
        # press another key
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(InputEvent.key(b, 1)))
        await asyncio.sleep(0.1)

        self.assertListEqual(self.result, [(EV_KEY, y, 1), (EV_KEY, y, 0)])
        self.assertFalse(macro.running)

    async def test_if_not_single_none(self):
        macro = Parser.parse("if_single(key(x),)", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 1)

        a = keyboard_layout.get("a")
        b = keyboard_layout.get("b")

        x = keyboard_layout.get("x")

        # press trigger key
        await self.trigger_sequence(macro, InputEvent.key(a, 1))
        await asyncio.sleep(0.1)
        # press another key
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(InputEvent.key(b, 1)))
        await asyncio.sleep(0.1)

        self.assertListEqual(self.result, [])
        self.assertFalse(macro.running)

    async def test_if_single_times_out(self):
        macro = Parser.parse(
            "set(t, 300).if_single(key(x), key(y), timeout=$t)",
            self.context,
            DummyMapping,
        )
        self.assertEqual(self.count_child_macros(macro), 2)

        a = keyboard_layout.get("a")
        y = keyboard_layout.get("y")

        await self.trigger_sequence(macro, InputEvent.key(a, 1))

        # no timeout yet
        await asyncio.sleep(0.2)
        self.assertListEqual(self.result, [])
        self.assertTrue(macro.running)

        # times out now
        await asyncio.sleep(0.2)
        self.assertListEqual(self.result, [(EV_KEY, y, 1), (EV_KEY, y, 0)])
        self.assertFalse(macro.running)

    async def test_if_single_ignores_joystick(self):
        """Triggers else + delayed_handle_keycode."""
        # Integration test style for if_single.
        # If a joystick that is mapped to a button is moved, if_single stops
        macro = Parser.parse(
            "if_single(k(a), k(KEY_LEFTSHIFT))", self.context, DummyMapping
        )
        code_shift = keyboard_layout.get("KEY_LEFTSHIFT")
        code_a = keyboard_layout.get("a")
        trigger = 1

        await self.trigger_sequence(macro, InputEvent.key(trigger, 1))
        await asyncio.sleep(0.1)
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(InputEvent.abs(ABS_Y, 10)))
        await asyncio.sleep(0.1)
        await self.release_sequence(macro, InputEvent.key(trigger, 0))
        await asyncio.sleep(0.1)
        self.assertFalse(macro.running)
        self.assertListEqual(self.result, [(EV_KEY, code_a, 1), (EV_KEY, code_a, 0)])

    async def test_raises_error(self):
        Parser.parse("if_single(k(a),)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "if_single(1,)", self.context)
        self.assertRaises(MacroError, Parser.parse, "if_single(,1)", self.context)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
import copy
import os
import subprocess
import time
from pickle import UnpicklingError
from unittest.mock import patch
import atexit

import evdev

from inputremapper.utils import get_evdev_constant_name
from tests.lib.constants import EVENT_READ_TIMEOUT, MIN_ABS, MAX_ABS
from tests.lib.fixtures import Fixture, fixtures, new_event
from tests.lib.pipes import (
    setup_pipe,
    push_events,
    uinput_write_history,
    uinput_write_history_pipe,
    pending_events,
)
from tests.lib.xmodmap import xmodmap
from tests.lib.tmp import tmp
from tests.lib.logger import logger


def patch_paths():
    from inputremapper.user import UserUtils
    from inputremapper.configs.paths import PathUtils

    return [
        patch.object(UserUtils, "home", tmp),
        patch.dict(os.environ, {"XDG_CONFIG_HOME": os.path.join(tmp, ".config")}),
    ]


class InputDevice:
    # expose as existing attribute, otherwise the patch for
    # evdev < 1.0.0 will crash the test
    path = None

    def __init__(self, path):
        if path != "justdoit" and not fixtures.get(path):
            # beware that fixtures keys and the path attribute of a fixture can
            # theoretically be different. I don't know if this is the case right now
            logger.error(
                'path "%s" was not found in fixtures. available: %s',
                path,
                list(fixtures.get_paths()),
            )
            raise FileNotFoundError()
        if path == "justdoit":
            self._fixture = Fixture(path="justdoit")
        else:
            self._fixture = fixtures[path]

        self.path = path
        self.phys = self._fixture.phys
        self.info = self._fixture.info
        self.name = self._fixture.name
        self.uniq = self._fixture.uniq

        # this property exists only for test purposes and is not part of
        # the original evdev.InputDevice class
        self.group_key = self._fixture.group_key or self._fixture.name

        # ensure a pipe exists to make this object act like
        # it is reading events from a device
        setup_pipe(self._fixture)

        self.fd = pending_events[self._fixture][1].fileno()

    def push_events(self, events):
        push_events(self._fixture, events)

    def fileno(self):
        """Compatibility to select.select."""
        return self.fd

    def log(self, key, msg):
        logger.info('%s "%s" "%s" %s', msg, self.name, self.path, key)

    def absinfo(self, *args):
        raise Exception("Ubuntus version of evdev doesn't support .absinfo")

    def grab(self):
        logger.info("grab %s %s", self.name, self.path)

    def ungrab(self):
        logger.info("ungrab %s %s", self.name, self.path)

    async def async_read_loop(self):
        logger.info("starting read loop for %s", self.path)
        new_frame = asyncio.Event()
        asyncio.get_running_loop().add_reader(self.fd, new_frame.set)
        while True:
            await new_frame.wait()
            new_frame.clear()
            if not pending_events[self._fixture][1].poll():
                # todo: why? why do we need this?
                # sometimes this happens, as if a other process calls recv on
                # the pipe
                continue

            event = pending_events[self._fixture][1].recv()
            logger.info("got %s at %s", event, self.path)
            yield event

    def read(self):
        # the patched fake InputDevice objects read anything pending from
        # that group.
        # To be realistic it would have to check if the provided
        # element is in its capabilities.
        if self.group_key not in pending_events:
            self.log("no events to read", self.group_key)
            return

        # consume all of them
        while pending_events[self._fixture][1].poll():
            event = pending_events[self._fixture][1].recv()
            self.log(event, "read")
            yield event
            time.sleep(EVENT_READ_TIMEOUT)

    def read_loop(self):
        """Endless loop that yields events."""
        while True:
            event = pending_events[self._fixture][1].recv()
            if event is not None:
                self.log(event, "read_loop")
                yield event
            time.sleep(EVENT_READ_TIMEOUT)

    def read_one(self):
        """Read one event or none if nothing available."""
        if not pending_events.get(self._fixture):
            return None

        if not pending_events[self._fixture][1].poll():
            return None

        try:
            event = pending_events[self._fixture][1].recv()
        except (UnpicklingError, EOFError):
            # failed in tests sometimes
            return None

        self.log(event, "read_one")
        return event

    def capabilities(self, absinfo=True, verbose=False):
        result = copy.deepcopy(self._fixture.capabilities)

        if absinfo and evdev.ecodes.EV_ABS in result:
            absinfo_obj = evdev.AbsInfo(
                value=None,
                min=MIN_ABS,
                fuzz=None,
                flat=None,
                resolution=None,
                max=MAX_ABS,
            )

            ev_abs = []
            for ev_code in result[evdev.ecodes.EV_ABS]:
                if ev_code in range(0x10, 0x18):  # ABS_HAT0X - ABS_HAT3Y
                    absinfo_obj = evdev.AbsInfo(
                        value=None,
                        min=-1,
                        fuzz=None,
                        flat=None,
                        resolution=None,
                        max=1,
                    )
                ev_abs.append((ev_code, absinfo_obj))

            result[evdev.ecodes.EV_ABS] = ev_abs

        return result

    def input_props(self):
        return []

    def leds(self):
        return []


uinputs = {}


class UInputMock:
    def __init__(self, events=None, name="unnamed", *args, **kwargs):
        self.fd = 0
        self.write_count = 0
        self.device = InputDevice("justdoit")
        self.name = name
        self.events = events
        self.write_history = []

        global uinputs
        uinputs[name] = self

    def capabilities(self, verbose=False, absinfo=True):
        if absinfo or 3 not in self.events:
            return self.events
        else:
            events = self.events.copy()
            events[3] = [code for code, _ in self.events[3]]
            return events

    def write(self, type, code, value):
        self.write_count += 1
        event = new_event(type, code, value, time.time())
        uinput_write_history.append(event)
        uinput_write_history_pipe[1].send(event)
        self.write_history.append(event)
        logger.info(
            '%s %s written to "%s"',
            (type, code, value),
            get_evdev_constant_name(type, code),
            self.name,
        )

    def syn(self):
        pass


def patch_evdev():
    def list_devices():
        return [fixture_.path for fixture_ in fixtures]

    class PatchedInputEvent(evdev.InputEvent):
        def __init__(self, sec, usec, type, code, value):
            self.t = (type, code, value)
            super().__init__(sec, usec, type, code, value)

        def copy(self):
            return PatchedInputEvent(
                self.sec,
                self.usec,
                self.type,
                self.code,
                self.value,
            )

    return [
        patch.object(evdev, "list_devices", list_devices),
        patch.object(evdev, "InputDevice", InputDevice),
        patch.object(evdev.UInput, "capabilities", UInputMock.capabilities),
        patch.object(evdev.UInput, "write", UInputMock.write),
        patch.object(evdev.UInput, "syn", UInputMock.syn),
        patch.object(evdev.UInput, "__init__", UInputMock.__init__),
        patch.object(evdev, "InputEvent", PatchedInputEvent),
    ]


def patch_events():
    # improve logging of stuff
    return patch.object(
        evdev.InputEvent,
        "__str__",
        lambda self: (f"InputEvent{(self.type, self.code, self.value)}"),
    )


def patch_os_system():
    """Avoid running pkexec."""
    original_system = os.system

    def system(command):
        if "pkexec" in command:
            # because it
            # - will open a window for user input
            # - has no knowledge of the fixtures and patches
            raise Exception("Write patches to avoid running pkexec stuff")
        return original_system(command)

    return patch.object(os, "system", system)


def patch_atexit_register():
    """Avoid adding tons of redundant atexit handlers that we don't need anyway.
    Otherwise we get lots of logs at the end of gui tests that bury the test result.
    """
    return patch.object(atexit, "register")


def patch_check_output():
    """Xmodmap -pke should always return a fixed set of symbols.

    On some installations the `xmodmap` command might be missig completely,
    which would break the tests.
    """
    original_check_output = subprocess.check_output

    def check_output(command, *args, **kwargs):
        if "xmodmap" in command and "-pke" in command:
            return xmodmap
        return original_check_output(command, *args, **kwargs)

    return patch.object(subprocess, "check_output", check_output)


def patch_regrab_timeout():
    # no need for a high number in tests
    from inputremapper.injection.injector import Injector

    return patch.object(Injector, "regrab_timeout", 0.05)


def is_running_patch():
    logger.info("is_running is patched to always return True")
    return True


def patch_is_running():
    from inputremapper.gui.reader_service import ReaderService

    return patch.object(ReaderService, "is_running", is_running_patch)


class FakeDaemonProxy:
    def __init__(self):
        self.calls = {
            "stop_injecting": [],
            "get_state": [],
            "start_injecting": [],
            "stop_all": 0,
            "set_config_dir": [],
            "autoload": 0,
            "autoload_single": [],
            "hello": [],
            "quit": 0,
        }

    def stop_injecting(self, group_key: str) -> None:
        self.calls["stop_injecting"].append(group_key)

    def get_state(self, group_key: str):
        from inputremapper.injection.injector import InjectorState

        self.calls["get_state"].append(group_key)
        return InjectorState.STOPPED

    def start_injecting(self, group_key: str, preset: str) -> bool:
        self.calls["start_injecting"].append((group_key, preset))
        return True

    def stop_all(self) -> None:
        self.calls["stop_all"] += 1

    def set_config_dir(self, config_dir: str) -> None:
        self.calls["set_config_dir"].append(config_dir)

    def autoload(self) -> None:
        self.calls["autoload"] += 1

    def autoload_single(self, group_key: str) -> None:
        self.calls["autoload_single"].append(group_key)

    def hello(self, out: str) -> str:
        self.calls["hello"].append(out)
        return out

    def quit(self):
        self.calls["quit"] += 1


def create_patches():
    return [
        # Sketchy, they only work because the whole modules are imported, instead of
        # importing `check_output` and `system` from the module.
        *patch_evdev(),
        patch_os_system(),
        patch_atexit_register(),
        patch_check_output(),
        # Those are comfortably wrapped in a class, and are therefore easy to patch
        *patch_paths(),
        patch_regrab_timeout(),
        patch_is_running(),
        patch_events(),
    ]



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import re
from dataclasses import dataclass
from typing import Dict, Tuple, Optional, Callable

from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.mapping import MappingData
from inputremapper.gui.messages.message_types import (
    MessageType,
    Name,
    Capabilities,
    Key,
    DeviceTypes,
)


@dataclass(frozen=True)
class UInputsData:
    message_type = MessageType.uinputs
    uinputs: Dict[Name, Capabilities]

    def __str__(self):
        string = f"{self.__class__.__name__}(uinputs={self.uinputs})"

        # find all sequences of comma+space separated numbers, and shorten them
        # to the first and last number
        all_matches = list(re.finditer(r"(\d+, )+", string))
        all_matches.reverse()
        for match in all_matches:
            start = match.start()
            end = match.end()
            start += string[start:].find(",") + 2
            if start == end:
                continue
            string = f"{string[:start]}... {string[end:]}"

        return string


@dataclass(frozen=True)
class GroupsData:
    """Message containing all available groups and their device types."""

    message_type = MessageType.groups
    groups: Dict[Key, DeviceTypes]


@dataclass(frozen=True)
class GroupData:
    """Message with the active group and available presets for the group."""

    message_type = MessageType.group
    group_key: str
    presets: Tuple[str, ...]


@dataclass(frozen=True)
class PresetData:
    """Message with the active preset name and mapping names/combinations."""

    message_type = MessageType.preset
    name: Optional[Name]
    mappings: Optional[Tuple[MappingData, ...]]
    autoload: bool = False


@dataclass(frozen=True)
class StatusData:
    """Message with the strings and id for the status bar."""

    message_type = MessageType.status_msg
    ctx_id: int
    msg: Optional[str] = None
    tooltip: Optional[str] = None


@dataclass(frozen=True)
class CombinationRecorded:
    """Message with the latest recoded combination."""

    message_type = MessageType.combination_recorded
    combination: "InputCombination"


@dataclass(frozen=True)
class CombinationUpdate:
    """Message with the old and new combination (hash for a mapping) when it changed."""

    message_type = MessageType.combination_update
    old_combination: "InputCombination"
    new_combination: "InputCombination"


@dataclass(frozen=True)
class UserConfirmRequest:
    """Message for requesting a user response (confirm/cancel) from the gui."""

    message_type = MessageType.user_confirm_request
    msg: str
    respond: Callable[[bool], None] = lambda _: None


@dataclass(frozen=True)
class DoStackSwitch:
    """Command the stack to switch to a different page."""

    message_type = MessageType.do_stack_switch
    page_index: int
